
short_name = "ACS"

long_name = "Advanced Camera for Surveys"

index_page_text = "Supported modes: imaging, ramp filter, spectroscopic "

modes = [
    ( 'imaging',        __name__ + '.web.imaging' ),
    ( 'spectroscopic',  __name__ + '.web.spectroscopic' ),
    ( 'rampfilter',     __name__ + '.web.rampfilter' ),
]

import pandeia.engine.helpers.bit.instruments.hst.default as default
mode_names = default.default_mode_names

short_mode_names = default.default_short_mode_names

import pandeia.engine.helpers.bit.instruments as pi
import pandeia.engine.helpers.bit.instruments.data

location=__file__

data = pi.data.Data('hst', 'acs', location=location)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Warning defined here are for conditions that can affect the results of a calculation, but
do not prevent a calculation from proceeding without error.

API warnings about the use of default values in calculations are handled programmatically
within each class.
"""

# these are warning messages that are common to all modules. the specific warnings
# currently defined will probably go away once API and sanity checking is fully
# implemented and verified throughout the system. these messages get combined with each of
# the module-specific messages using the update() method.
standard_warning_messages = {
    "no_sanity_check": "No sanity checking implemented for %s: %s",
    "no_api_check": "No API checking implemented for %s: %s",
    "missing_api": "API parameter, {:s}, missing from input for {:s}. Using default value of {}.",
    "unsupported_api": "Unsupported configuration parameter, {:s}, being passed to {:s}."
}

# warning messages specific to signal and its classes
signal_warning_messages = {
    "background_saturated": "Background is saturating. Consider making the observation shorter.",
    "brightest_pixel_wavelength_centered": "Brightest pixel wavelengths assume that the scene is a single centered source. If the target is not centered, or the brightest pixel is not in the centered target spectrum, the wavelength will be incorrect."
}

# warning messages for the instrument class
instrument_warning_messages = {
    "missing_instrument_api": "Missing {:s} API parameter {:s} for {:s}. Using default value of {:s}.",
    "spectral_resampling_too_small": "Spectrum for source {:s} has been resampled to a spacing " \
                                 "of {} microns for performance reasons. This is smaller " \
                                 "than the minimum spectral resolution {}.",
    "roman_max_resultants": "Using maximum number of resultants for this MA table ({})"
}

source_warning_messages = {
    "source_parameters_missing": "Missing list of required API parameters for source {:s}.",
    "source_parameter_missing": "Source {} configuration missing API parameter, {}. Using the default value of {}."
}


# warning messages specific to Strategy and its sub-classes.
strategy_warning_messages = {
    "background_region_too_small": "Background region smaller than source extraction region. This can adversely affect the SNR.",
    "extraction_aperture_truncated": "Extraction aperture partially outside of the field of view.",
    "background_region_truncated": "Background estimation region partially outside of the field of view.",
    "extraction_aperture_undersampled": "Extraction aperture undersampled. Pixel area %.2f%% less than requested area.",
    "background_region_undersampled": "Background estimation region undersampled. Pixel area %.2f%% less than requested area.",
    "upper_background_region_missing": "Upper background estimation region outside of field of view.",
    "upper_background_region_truncated": "Upper background estimation region truncated by %.2f%%.",
    "lower_background_region_missing": "Lower background estimation region outside of field of view.",
    "lower_background_region_truncated": "Lower background estimation region truncated by %.2f%%.",
    "target_occulted": "Specified target position is occulted by the coronagraphy mask.",
    "coronagraphy_central_source": "Coronagraphy requires a central source within {:.2f} of the center.",
    "coronagraphy_psf_source": "Coronagraphy requires the PSF subtraction source to be outside the scene (>{:.2f} arcsec, not {:.2f}",
    "strategy_unsupported_soss_bknd": "Background subtraction not yet implemented for SOSS.",
    "ccd_imageacq_slit": "Slit {} for acq_mode {} is unsupported."
}
strategy_warning_messages.update(standard_warning_messages)

# warning messages specific to Normalization and its sub-classes
normalization_warning_messages = {
    "normalized_to_zero_flux": "Zero flux at reference wavelength. Spectrum left unscaled.",
    "normalized_to_zero_flux_bandpass": "Zero flux in bandpass. Spectrum left unscaled.",
    "unsupported_normalization_bandpass": "Bandpass specification, %s, not currently supported, but may work."
}
normalization_warning_messages.update(standard_warning_messages)

# warning messages specific to SED and its sub-classes. currently none defined...
sed_warning_messages = {}
sed_warning_messages.update(standard_warning_messages)

# warning messages specific to Telescope and its sub-classes
telescope_warning_messages = {
    "telescope_ote_efficiency_missing": "Telescope OTE throughput mis-configured or unavailable. Using default value of 1.0.",
    "telescope_background_missing": "Telescope notional background mis-configured or unavailable. Using default value of 0.0."
}
telescope_warning_messages.update(standard_warning_messages)

# warning messages specific to AstroSpectrum and its classes and methods
astrospectrum_warning_messages = {
    "max_scene_size_reached": "Scene requires a total field-of-view of %.3f arcsec. Using the configured maximum of %.3f arcsec.",
    "scene_fov_too_small": "Field-of-view size, %.3f, too small to encompass any of the defined sources.",
    "wavelength_truncated_blue": "Spectrum blue limit, %.2f, does not extend to instrument configuration blue limit, %.2f.",
    "wavelength_truncated_red": "Spectrum red limit, %.2f, does not extend to instrument configuration red limit, %.2f.",
    "scene_range_truncated": "Combined wavelength range of scene [%.2f, %.2f] less than instrument configuration's range [%.2f, %.2f].",
    "spectrum_missing_blue": "Spectrum [%.2f, %.2f] does not extend to instrument configuration blue limit, %.2f.",
    "spectrum_missing_red": "Spectrum [%.2f, %.2f] does not extend to instrument configuration red limit, %.2f.",
    "filter_leak": "This observation has a significant filter leak. {0:.3f} percent of the signal of one of the sources comes from outside the prescribed filter bandpass limits and is not included in the results.",
}
astrospectrum_warning_messages.update(standard_warning_messages)

report_warning_messages = {
    "bad_waveref": "Specified wavelength, {0:.2f}, out of range [{1:.2f}, {2:.2f}]. Using {3:.2f} to select diagnostic planes instead.",
    "no_waveref": "No specified wavelength for spectral mode. Using {0:.2f} to select diagnostic planes.",
    "cos_extended_multiple": "COS is not intended for use with multiple sources, offset sources, or extended sources.",
    "bop_single_centered_source": "Bright Object Protection values are only accurate for scenes with a single, centered source.  "\
    "<a target=\"_blank\" href=\"https://hst-docs.stsci.edu/hetc/calculations-page-overview/saturation-and-bright-object-limits\">Please consult the documentation here.</a>"
}
report_warning_messages.update(standard_warning_messages)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

from . import source as src
from .custom_exceptions import EngineInputError


class Scene(object):

    """
    Container class describing a scene as composed of a list of Sources.

    Parameters
    ----------
    config : dict or list (optional, default None)
        Dictionary input must contain a 'Scene' section concordant with the engine input API.
        Can also provide a pre-made list of source.Source objects. If no input supplied,
        then default to single default source.Source().
    telescope : str
        The name of the telescope for which the scene is being created.

    Attributes
    ----------
    sources : list
        The list of sources that compose the scene
    offset : list-like (optional)
        Offset to apply to sources upon instantiation given as [x, y]

    Methods
    -------
    offset(x=<float, default 0.0>, y=<float, default 0.0>) :
        Apply x and/or y offset to all sources in scene
    get_size():
        Get the full extent of the scene.
    """

    def __init__(self, telescope, config=None, offset=None, webapp=False, **kwargs):
        self.webapp = webapp
        self.warnings = {}
        self.telescope = telescope
        self.sources = []
        if isinstance(config, dict):
            try:
                # if we get a full engine API dict, fish the scene out
                if 'scene' in config['scene']:
                    self._from_list(config['scene'])
                else:
                    # if we're a dict, but not a full engine API one, try adding as a source
                    self.add_sources(config)
            except KeyError as e:
                message = "Input dict must have source information in engine API format: %s" % e
                raise EngineInputError(value=message)
        elif isinstance(config, list):
            self._from_list(config)
        elif not self.webapp:
            # add the default source if there is none, and webapp==False
            self.sources = [src.Source(self.telescope)]
        else:
            msg = "Must provide scene configuration."
            raise EngineInputError(value=msg)

        if offset is not None:
            if isinstance(offset, list) or isinstance(offset, tuple):
                if len(offset) == 2:
                    self.offset({'x': offset[0], 'y': offset[1]})
                elif isinstance(offset, dict):
                    self.offset(offset)
                else:
                    message = "Offset needs to be list-like of format [x, y] "
                    message += "or a dict of format {'x': x, 'y': y}"
                    raise EngineInputError(value=message)
            else:
                message = "If specified, offset needs to be list-like of format [x, y] "
                message += "or a dict of format {'x': x, 'y': y}."
                raise EngineInputError(value=message)

        for s in self.sources:
            self.warnings.update(s.warnings)

    def _from_list(self, source_list):
        """
        Add sources to a scene from a list. Can consist of src.Source() instances or dicts that
        can be used to instantiate a src.Source() or a mixture of the two.

        Parameters
        ----------
        source_list: list
            List consisting of src.Source() instances or dicts for configuring a src.Source()
        """
        for s in source_list:
            if isinstance(s, src.Source):
                self.sources.append(s)
            if isinstance(s, dict):
                self.sources.append(src.Source(self.telescope, config=s, webapp=self.webapp))

    def add_sources(self, s):
        """
        Add a source to a scene. Can be either src.Source instance or a dict containing
        configuration information for a src.Source.

        Parameters
        ----------
        s: src.Source or dict
            A source instance or configuration information for creating source
        """
        if isinstance(s, src.Source):
            self.sources.append(s)
        elif isinstance(s, dict):
            self.sources.append(src.Source(self.telescope, config=s, webapp=self.webapp))
        elif isinstance(s, list):
            self._from_list(s)
        else:
            message = "Must provide either Source instance(s) or a configuration dict(s) to create them."
            raise EngineInputError(value=message)

    def offset(self, dither):
        """
        Apply an offset to all sources within the Scene instance.

        Parameters
        ----------
        dither : dict - {'x': x, 'y': y}
            x : float (default 0.0)
                offset in the X direction
            y : float (default 0.0)
                offset in the Y direction
        """
        for i, s in enumerate(self.sources):
            self.sources[i].position.update(
                {
                    'x_offset': s.position['x_offset'] + dither['x'],
                    'y_offset': s.position['y_offset'] + dither['y']
                }
            )

    def set_pixsamp(self, xsamp=1.0, ysamp=1.0):
        """
        Set the pixel sampling to apply to the scene

        Parameters
        ----------
        xsamp: float
            Pixel size along the X axis
        ysamp: float
            Pixel size along the Y axis
        """
        for s in self.sources:
            s.shape['xsamp'] = xsamp
            s.shape['ysamp'] = ysamp

    def get_size(self):
        """
        Get the relative extent of the scene based on the location of the sources in it.

        Parameters
        ----------
        None

        Returns
        -------
        size: float
            Size of the scene
        """

        # The scene size is the maximum (or minimum) x or y offset.
        # It is multiplied by 2 because the 0 position is referenced to the center of the FOV.
        size = 2.0 * np.max([np.max(np.abs([s.position['x_offset'], s.position['y_offset']])) for s in self.sources])

        return size

    def get_min_size(self):
        """
        Get the minimum scene size required to encompass at least one of the sources.

        Parameters
        ----------
        None

        Returns
        -------
        min_size: float
            Minimum size of the scene
        """

        # The scene size is the maximum (or minimum) x or y offset.
        # It is multiplied by 2 because the 0 position is referenced to the center of the FOV.
        min_size = 2.0 * np.min([np.max(np.abs([s.position['x_offset'], s.position['y_offset']])) for s in self.sources])

        return min_size



""" WFC3UVIS-specific functions

"""



from .. import scan_mode

from ....engine.imager import Imager
from ....engine.string_constants import RECTANGLE


def make_snr_region(instrument, *args, **kwargs):
    """
    WFC3UVIS imaging in scan mode uses a specialized SNR
    rectangular region.

    """
    if instrument.scan_mode:
        return scan_mode.make_rectangular_img_snr_region(args, instrument)
    else:
        return None


def get_configuration_data(ei):
    """ Load instrument and detector data from config files.

        This overrides the base function in order to handle
        WFC3UVIS specifics.

    Parameters
    ----------
    ei: dict
      engine input

    Returns
    -------
    iconfig: dict
      dictionary of instrument data
    dconfig: dict
      dictionary of detector data
    custom_config: dict
      dictionary with custom data

    """
    # get baseline configuration data.
    return scan_mode.get_scan_configuration_data(ei)


def modify_configuration(instrument, custom_config):
    """ The scan_mode flag must be propagated into the instrument
        instance, and also from the raw table data into instances
        of EE table.

    Parameters
    ----------
    instrument: Imager or Spectrograph
      Instance being configured
    custom_config: dictionary
      custom configuration info

    """
    scan_mode.modify_scan_configuration(instrument, custom_config)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

from .coords import Grid, IrregularGrid
from .utils import recursive_subclasses
from .custom_exceptions import EngineOutputError, EngineInputError

def ProjectionFactory(projection_type):
    """
    Function to take projection type and return the appropriate projection-dependent functions.

    Parameters
    ----------
    projection_type: string
        String with six valid values: "image", "image_scan", "spec", "slitless", 
        "slitless_scan", "multiorder".
    """

    types = recursive_subclasses(Projection)
    projections = [t.__name__.lower().replace('projection', '') for t in types]
    type_map = dict(list(zip(projections, types)))

    if projection_type not in projections:
        msg = "Unsupported or not yet implemented projection: %s" % projection_type
        raise EngineInputError(value=msg)
    else:
        cls = type_map[projection_type]()
        return cls


class Projection:
    """
    Abstract class with projection-type dependent code. Currently only includes 
    projection-dependent aspects of the Report class, but should eventually be expanded to 
    include projection-dependent parts of signal.py, noise.py, and possibly instrument.py

    Note that none of these correspond to the actual 2D results of the IFU projection type, 
    which technically isn't a projection as it reports the full scene cubes.

    It also does not include Coronagraphy, which (uniquely) has no background to subtract 
    from its _2d results.

    """

    def __init__(self):
        pass

    def _2d(self, extracted, saturation, signal):

        self.signal = signal

        # Signal and noise in 2D. Get from extracted products
        s = extracted['detector_signal']
        n = extracted['detector_noise']
        t = extracted['detector_saturation']

        # This is the background rate in each pixel without sources
        self.bg_pix = signal.rate_plus_bg - signal.rate

        # The noise is undefined if the pixel is fully saturated.
        # The engine currently returns noise=0 if the pixel has full saturation, which is
        # confusing since the noise is not 0, but rather undetermined or infinite. Setting
        # the noise to NaN ensures that the S/N of saturated pixels are NaNs.
        n[t == 2] = np.nan
        self.detector_sn = (s - self.bg_pix) / n
        self.detector_signal = s + n * np.random.randn(n.shape[0], n.shape[1])
        self.saturation = t
        self.ngroups_map = extracted['detector_ngroups']


class ImageProjection(Projection):

    def _2d(self, extracted, saturation, signal):
        Projection._2d(self,extracted,saturation,signal)

        return {'detector': self.detector_signal, 'snr': self.detector_sn, 'saturation': self.saturation, \
                'ngroups_map': self.ngroups_map}, self.bg_pix

    def _project(self, signal, extracted):

        return signal.grid, signal.wave_pix, signal.pixgrid_list[0]

class Image_ScanProjection(ImageProjection):
    pass
class SpecProjection(Projection):

    def _2d(self, extracted, saturation, signal):
        Projection._2d(self,extracted,saturation,signal)

        if self.signal.dispersion_axis == 'y' and self.signal.current_instrument.instrument["disperser"] != "p750l":  
            # where projection is slitless and axis is y (and not the MIRI p750l disperser)
            # need to rotate these 90 deg clockwise to match our normal axis orientation. np.rot90 only works CCW
            # so we need to rotate 3 times.
            det_sn = np.rot90(self.detector_sn, 3)
            det_signal = np.rot90(self.detector_signal, 3)
            det_sat = np.rot90(self.saturation, 3)
            det_groups = np.rot90(self.ngroups_map, 3)
        else:
            det_signal = self.detector_signal[::-1,:]
            det_sn = self.detector_sn[::-1,:]
            det_sat = self.saturation[::-1,:]
            det_groups = self.ngroups_map[::-1,:]

        return {'detector_unrotated': self.detector_signal, 'snr_unrotated': self.detector_sn, \
                'saturation_unrotated': self.saturation, 'ngroups_map_unrotated': self.ngroups_map, \
                'detector':det_signal, 'snr':det_sn, 'saturation':det_sat, 'ngroups_map': det_groups}, \
               self.bg_pix

    def _project(self, signal, extracted):

        return signal.grid, extracted['wavelength'], signal.pixgrid_list[0]


class SlitlessProjection(Projection):

    def _2d(self, extracted, saturation, signal):

        Projection._2d(self,extracted,saturation,signal)

        if self.signal.dispersion_axis == 'y':
            if self.signal.current_instrument.instrument["disperser"] == "p750l":
                # y-dispersed spectra need to be flipped vertically.
                det_signal = self.detector_signal[::-1,:]
                det_sn = self.detector_sn[::-1,:]
                det_sat = self.saturation[::-1,:]
                det_groups = self.ngroups_map[::-1,:]
            else:
                # where projection is slitless and axis is y (and not the MIRI p750l disperser)
                # need to rotate these 90 deg clockwise to match our normal axis orientation. np.rot90 only works CCW
                # so we need to rotate 3 times.
                det_sn = np.rot90(self.detector_sn, 3)
                det_signal = np.rot90(self.detector_signal, 3)
                det_sat = np.rot90(self.saturation, 3)
                det_groups = np.rot90(self.ngroups_map, 3)
        else:
            det_signal = self.detector_signal
            det_sn = self.detector_sn
            det_sat = self.saturation
            det_groups = self.ngroups_map

        return {'detector_unrotated': self.detector_signal, 'snr_unrotated': self.detector_sn, \
                'saturation_unrotated': self.saturation, 'ngroups_map_unrotated': self.ngroups_map, \
                'detector':det_signal, 'snr':det_sn, 'saturation':det_sat, 'ngroups_map': det_groups}, \
               self.bg_pix

    def _project(self, signal, extracted):
        '''
        This function handles (almost) all projection-type dependent factors (which are independent of strategy)
        This primarily includes wave_pix and pix_grid.

        The 2D projection types for slitless and multiorder are handled in _2d()
        '''
        # this is the spatial grid for the calculation. It needs to be added as an attribute for the
        # common/scene/coordinates tests to work.
        grid = signal.grid
        wave_pix = extracted['wavelength']
        if signal.dispersion_axis == 'y':
            wave_pix = wave_pix[::-1]
        # this is the detector plane pixel grid. for most modes we just grab and use it directly.
        # however, slitless we want to redefine it to be spatial on both axes.
        orig_grid = signal.pixgrid_list[0]
        # detector plane gets rotated depending on dispersion_axis so adjust accordingly
        if signal.dispersion_axis == 'x' or signal.current_instrument.instrument["disperser"] == "p750l":
            pix_grid = Grid(grid.xsamp, orig_grid.ysamp, orig_grid.nx, orig_grid.ny)
        else:
            pix_grid = Grid(grid.ysamp, orig_grid.xsamp, orig_grid.ny, orig_grid.nx)

        return grid, wave_pix, pix_grid

class Slitless_ScanProjection(SlitlessProjection):
    pass

class MultiorderProjection(SlitlessProjection):

    def _2d(self, extracted, saturation, signal):
        Projection._2d(self,extracted,saturation,signal)
        
        return {'detector_unrotated': self.detector_signal, 'snr_unrotated': self.detector_sn, \
                'saturation_unrotated': self.saturation, 'ngroups_map_unrotated': self.ngroups_map, \
                'detector':np.rot90(self.detector_signal), 'snr':np.rot90(self.detector_sn),
                'saturation':np.rot90(self.saturation), 'ngroups_map': np.rot90(self.ngroups_map)}, self.bg_pix

    def _project(self, signal, extracted):
        '''
        This function handles (almost) all projection-type dependent factors (which are independent of strategy)
        This primarily includes wave_pix and pix_grid.

        The 2D projection types for slitless and multiorder are handled in _2d()
        '''
        # this is the spatial grid for the calculation. It needs to be added as an attribute for the
        # common/scene/coordinates tests to work.
        grid = signal.grid
        # this is the wavelength sampling on the detector.
        wave_pix = extracted['wavelength']  # this is already a 1D np.array

        orig_grid = signal.pixgrid_list[0]
        pix_grid = IrregularGrid(np.arange(orig_grid.nx,0,-1), np.arange(orig_grid.ny))

        return grid, wave_pix, pix_grid



"""
Functions converted from C to python from HEALPix file
chealpix.c

"""
import math

twopi = 2. * math.pi
inv_halfpi = 2. / math.pi
twothird = 2. / 3.


# /*! Returns the remainder of the division \a v1/v2.
#   The result is non-negative.
#    \a v1 can be positive or negative; \a v2 must be positive. */

def fmodulo(v1, v2):
    if v1 >= 0.:
        if (v1 < v2):
            return v1
        else:
            return math.fmod(v1, v2)
    tmp = math.fmod(v1, v2) + v2
    if tmp == v2:
        return 0.
    else:
        return tmp
# /*  return (v1>=0) ? ((v1<v2) ? v1 : fmod(v1,v2)) : (fmod(v1,v2)+v2); */
# 3  }
# *! Returns the remainder of the division \a v1/v2.
#    The result is non-negative.
#    \a v1 can be positive or negative; \a v2 must be positive. */


def imodulo(v1, v2):
    v1 = int(v1)
    v2 = int(v2)
    v = v1 % v2
    if (v >= 0):
        return v
    else:
        return v + v2

# static int isqrt(int v)
#  { return (int)(sqrt(v+0.5)); }


def ang2pix_ring(nside, ra, dec):
    """Converts RA and Dec to HEALPix pixel index.

    Parameters
    ----------
    ra : double
        Right Ascension [radians]
    dec : double
        declination [radians]
    nside : int
        HEALPix scale factor.

    Returns
    -------
    ipix : int
        HEALPix pixel index.
    """

    theta = math.pi / 2. - dec
    phi = ra
    if theta < 0. or theta > math.pi:
        print("theta out of range in ang2pix_ring")
    return ang2pix_ring_z_phi(nside, math.cos(theta), phi)


def ang2pix_ring_z_phi(nside, z, phi):
    za = abs(z)
    tt = fmodulo(phi, twopi) * inv_halfpi  # in [0,4)

    if za <= twothird:  # Equatorial region
        temp1 = nside * (0.5 + tt)
        temp2 = nside * z * 0.75
        jp = int(temp1 - temp2)  # index of  ascending edge line
        jm = int(temp1 + temp2)  # index of descending edge line

        # ring number counted from z=2/3
        ir = nside + 1 + jp - jm  # ; /* in {1,2n+1} */
        kshift = 1 - (ir & 1)  # ; /* kshift=1 if ir even, 0 otherwise */  CHECK THIS &?

        ip = (jp + jm - nside + kshift + 1) / 2  # ; /* in {0,4n-1} */
        ip = imodulo(ip, 4 * nside)

        return nside * (nside - 1) * 2 + (ir - 1) * 4 * nside + ip
    else:  # /* North & South polar caps */
        tp = tt - (int)(tt)  # ;
        tmp = nside * math.sqrt(3 * (1 - za))  # ;

        jp = int(tp * tmp)  # ; /* increasing edge line index */
        jm = int((1.0 - tp) * tmp)  # ; /* decreasing edge line index */

        ir = jp + jm + 1  # ; /* ring number counted from the closest pole */
        ip = int(tt * ir)  # ; /* in {0,4*ir-1} */
        ip = imodulo(ip, 4 * ir)  # ;

        if z > 0.:
            return 2 * ir * (ir - 1) + ip
        else:
            return 12 * nside * nside - 2 * ir * (ir + 1) + ip



import datetime, os

from pandeia.engine.helpers.background.hst.util import read_dict
from pandeia.engine.helpers.background.hst import sky_angles as angles
from pandeia.engine.custom_exceptions import RangeError

'''
keep the synphot expression generation section of the code in sync with 
https://github.com/spacetelescope/pyetc/blob/master/pyetc/etc_web/etc/sky.py

keep everything except Sky class, formkeys, fd, descr, etype, ztype, level keys and error types in sync with 
https://github.com/spacetelescope/pandeia/blob/master/engine/pandeia/engine/helpers/background/hst/sky.py
'''
#TODO: sky.dat needs to migrate to CDBS & have a versioned name
MODELS = read_dict(os.path.join(os.path.dirname(__file__), '../background/hst/sky.dat'))

class SkyElement(object):
    """Base class for actual elements of the sky model"""
    #Subclasses will override this value
    formkeys = []


    def __init__(self, formdata=None):
        """formdata=a dictionary containing the key/value pairs required to
        instantiate this SkyElement
        """
        self.fd=dict()
        self.descr='Not yet set'
        self.expr='Not yet set'

        if formdata:
            #pick out what we care about
            for k in self.formkeys:
                self.fd[k]=formdata[k]
                #This will raise a KeyError if any keys of interest
                #are not present. The caller (typically the Sky constructor)
                #will have to deal with this exception

    def __str__(self):
        return str(self.descr)


    def make_synexpr(self):
        """Subclasses must implement this method. It should set both the
        .expr and the .descr attributes
        """
        raise NotImplementedError("Sorry, this must be implemented in the subclass")


class Airglow(SkyElement):
    """Airglow is modeled as a set of narrow emission lines in the UV."""

    #The only input values we care about
    formkeys = ['AirglowStandard']
    lines = ['el1215a',
             'el1302a',
             'el1356a',
             'el2471a']

    #Normalization constants for each level
    AIRGLOW_LEVELS = {
        'None': None,
        'Low': {'el1215a': 0.20,
                'el1302a': 0.01333333333,
                'el1356a': 0.012,
                'el2471a': 0.010},
        'Average': 1.0,
        'High': 2.0
        }


    def make_synexpr(self):
        self.descr=self.fd['AirglowStandard']
        if self.descr == 'None':
            self.expr = None
            return

        elif self.descr == 'Low':
            elements=[]
            #Special case, different multipliers for each line
            norm = self.AIRGLOW_LEVELS[self.descr]
            for l in self.lines:
                elements.append("%f*spec(%s)"%(norm[l],MODELS[l]))
            self.expr = ' + '.join(elements)

        else: #All others have a single multiplier applied to all lines
            elements = []
            for l in self.lines:
                elements.append("spec(%s)"%MODELS[l])
            self.expr = '(%s)'%' + '.join(elements)
            if self.AIRGLOW_LEVELS[self.descr] != 1.0:
                self.expr = "%f*(%s)"%(self.AIRGLOW_LEVELS[self.descr], self.expr)


class Earthshine(SkyElement):
    """Earthshine is modelled by a spectrum file."""

    formkeys = ['EarthshineStandard','EarthshineMag','EarthshineMult','EarthshineSpec']
    specfile = "spec(%s)"%MODELS['earthshine']
    # These are scalar multipliers
    EARTHSHINE_LEVELS = {
        'None': None,
        'Low': None,
        'Average': 0.5,
        'High': 1.0,
        'Extremely_high': 2.
        }

    def __init__(self, formdata=None):
        #Call the parent
        SkyElement.__init__(self, formdata)
        #This tells us which type of calculation we're going to do
        self.etype = self.fd['EarthshineSpec']
        self._set_norm()


    def make_synexpr(self):
        if self.norm is None:
            self.expr = None
        else:
            if self.etype in ('EarthshineMag'):
                self.expr = "rn(%s,band(johnson,v),%f,vegamag)"%(self.specfile,self.norm)
            elif self.etype in ('EarthshineStandard', 'EarthshineMult'):
                self.expr = self.specfile
                if self.norm != 1.0:
                    self.expr = "%f*%s"%(self.norm, self.expr)
            else:
                raise ValueError("Unrecognized normalization option %s"%self.etype)

    def _set_norm(self):
        if self.etype == 'EarthshineStandard':
            level = self.fd[self.etype]
            self.norm = self.EARTHSHINE_LEVELS[level]
            self.descr = level

        elif self.etype == 'EarthshineMag':
            self.norm = self.fd[self.etype]
            self.descr = 'User Defined (normalized to %f V)'%self.norm

        elif self.etype == 'EarthshineMult':
            self.norm = self.fd[self.etype]
            self.descr = 'User Defined (scaled by %f)'%self.norm

        else:
            raise ValueError("Unrecognized normalization option %s"%self.etype)


class Helium(SkyElement):
    """Helium is modelled by a spectrum file."""

    formkeys = ['HeliumStandard']
    specfile = "spec(%s)"%MODELS['helium']

    # These are scalar multipliers
    HELIUM_EMISSION_LEVELS = {
        'none':      None,
        'average':   0.1,
        'high':      0.5,
        'very_high': 1.5,
        }

    def __init__(self, formdata=None):
        SkyElement.__init__(self, formdata)
        self.descr = self.fd['HeliumStandard']
        self.norm = self.HELIUM_EMISSION_LEVELS[self.descr]

    def make_synexpr(self):
        if self.norm is None:
            self.expr = None
        else:
            self.expr = self.specfile
            if self.norm != 1.0:
                self.expr = "%f*%s"%(self.norm, self.expr)


class ZodiacalLight(SkyElement):
    """Zodiacal light is modeled by a spectrum file that can be
    normalized in a variety of ways, including based on the angle
    between the sun and the observed coordinates.
    """
    #TODO - Although only a subset of these keys are necessary to
    #construct a valid ZodiacalLight object, at present all are
    #required. We might override the __init__ completely and make it
    #complicated enough to cope with conditionals here.
    formkeys = ['ZodiSpec','ZodiStandard','ZodiRA','ZodiDec',
                'ZodiSunAttribute', 'ZodiSun', 'ZodiMag', 'ZodiMult',
                'ZodiDate', 'ZodiMonth', 'ZodiDay']
    specfile = "spec(%s)"%MODELS['zodi']
    ZODI_LEVELS = {  #Normalized to Johnson V
        'None': None,
        'Low': 23.3,
        'Average': 22.7,
        'High': 22.1
}

    #TODO: handle the posfile in the same way as the other data files
    #  except that some variable expansion has to be done, unlike for
    #  the other files for which that is handled by pysynphot
    #posfile = MODELS['zodipos']

    def __init__(self, formdata=None):
        #Then call the parent
        SkyElement.__init__(self, formdata)
        #This tells us which type of calculation we're going to do
        self.ztype = self.fd['ZodiSpec']
        self._set_norm()

    def make_synexpr(self):
        if self.norm is None:
            self.expr = None
        else:
            if self.ztype in ('ZodiStandard', 'ZodiMag', 'ZodiCoords'):
                self.expr = "rn(%s,band(johnson,v),%f,vegamag)"%(self.specfile,self.norm)
            elif self.ztype in ('ZodiMult'):
                self.expr = "%f*%s"%(self.norm,self.specfile)

    def _set_norm(self):
        #There are many ways to specify how the Zodiacal Light spectrum should be normalized
        if self.ztype == 'ZodiStandard':
            level = self.fd[self.ztype]
            self.norm = self.ZODI_LEVELS[level]
            self.descr = level

        elif self.ztype == 'ZodiMag':
            self.norm = self.fd[self.ztype]
            self.descr = 'User Defined (normalized to %f V)'%self.norm

        elif self.ztype == 'ZodiMult':
            self.norm = self.fd[self.ztype]
            self.descr = 'User Defined (scaled by %f)'%self.norm

        elif self.ztype == 'ZodiCoords':
            zra, zdec = self.fd['ZodiRA'], self.fd['ZodiDec']
            ra, dec = angles.radec_to_decimal( zra, zdec )

            if self.fd['ZodiSunAttribute'] == 'Date':
                year, month, day = self.fd['ZodiDate'], self.fd['ZodiMonth'], self.fd['ZodiDay']
                date = datetime.date(year,month,day)
                helong, helat, sunangle = angles.radec_date_to_helonglat(ra, dec, date)
                self.descr = ('%s %s %s/%s/%s (%f deg.)' % (zra,zdec,month,day,year,sunangle))

            elif self.fd['ZodiSunAttribute'] == 'SunAngle':
                zsun = self.fd['ZodiSun']
                # Table is defined from 0-180, but users can put in anything
                sunangle = angles.limit_longitude(float(zsun))
                self.descr = ('%s %s Sun Angle = %f deg.' % (zra, zdec, zsun))
                helong, helat = angles.radec_sunangle_to_helonglat(ra, dec, sunangle)

            else:
                raise ValueError('Unknown sun specification %s'%self.fd['ZodiSunAttribute'])
            try:
                self.norm = angles.zodi_lookup(helong, helat)
            except RangeError as e:
                e.addinfo("These values are derived from the RA, Dec, and Date or Helio-ecliptic Longitude. You must either select different values for these quantities, or a different normalization method for the zodiacal light.")
                raise

        else:
            raise ValueError('Unknown normalization type %s' % self.ztype)


class Sky(object):
    """Class to compute the sky background including all relevant SkyElements.
    Presently supports these elements:
        Airglow, Earthshine, Helium, and ZodiacalLight.
    Not all the elements are required. Elements will be picked out based on
    the contents of the dictionary that is passed in."""
    def __init__(self, formdata=None):
        self.airglow=None
        self.earthshine=None
        self.helium=None
        self.zodi=None
        self.descr=dict()
        self.expr='Not yet set'

        if formdata:
            try:
                self.airglow=Airglow(formdata)
                self.airglow.make_synexpr()
                self.descr['airglow_str']=self.airglow.descr
            except KeyError as e:
                pass

            try:
                self.earthshine=Earthshine(formdata)
                self.earthshine.make_synexpr()
                self.descr['earthshine_str']=self.earthshine.descr
            except KeyError as e:
                pass
            try:
                self.helium=Helium(formdata)
                self.helium.make_synexpr()
                self.descr['helium_str']=self.helium.descr
            except KeyError as e:
                pass

            try:
                self.zodi= ZodiacalLight(formdata)
                self.zodi.make_synexpr()
                self.descr['zodi_str']=self.zodi.descr
            except KeyError as e:
                pass

    def __str__(self):
        return str(self.descr)

    def make_synexpr(self):
        elements = []
        for x in (self.airglow, self.earthshine, self.helium, self.zodi):
            if x and x.expr:
                elements.append(x.expr)
        if len(elements) > 0:
            self.expr = ' + '.join(elements)
        else:
            #Provide a valid expression with a value of zero
            self.expr = 'unit(0.0,flam)'



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

from astropy.io import fits

import numpy.ma as ma
import numpy as np

from .telescope import Telescope
from .instrument import Instrument
from .io_utils import read_json
from .psf_library import PSFLibrary
from .custom_exceptions import EngineInputError, RangeError, DataError

class JWST(Telescope):

    """
    This is currently a dummy class that is used for configuration file discovery. Could eventually
    contain JWST-specific methods.
    """
    pass


class JWSTInstrument(Instrument):

    """
    Generic JWST Instrument class
    """

    def __init__(self, mode=None, config={}, webapp=False, **kwargs):
        telescope = JWST()
        # these are the required sections and need to be passed via API in webapp mode
        self.instrument_pars = {}
        self.instrument_pars['detector'] = ["nexp", "ngroup", "nint", "readout_pattern", "subarray"]
        self.instrument_pars['instrument'] = ["aperture", "disperser", "filter", "instrument", "mode"]
        self.api_parameters = list(self.instrument_pars.keys())

        # these are required for calculation, but ok to live with config file defaults
        if hasattr(self, 'api_ignore'):
            self.api_ignore.extend(['dynamic_scene', 'max_scene_size', 'scene_size'])
        else:
            self.api_ignore = ['dynamic_scene', 'max_scene_size', 'scene_size']

        Instrument.__init__(self, telescope=telescope, mode=mode, config=config, webapp=webapp, **kwargs)

    def read_detector(self):
        """Read in the detector keyword from the aperture parameter in the config json file.
           Put the detector keyword in self.instrument['detector']."""
        if "slit" in self.aperture_config[self.get_aperture()]:
            self.instrument['slit'] = self.aperture_config[self.get_aperture()]['slit']
        if "detector" not in self.instrument:
            self.instrument['detector'] = self.aperture_config[self.get_aperture()]['detector']

class NIRSpec(JWSTInstrument):

    """
    Need to over-ride get_wave_range() for NIRSpec because the effective wavelength range
    depends on the blocking filter, the aperture, and the disperser.  Also need to overload __init__
    to handle some special MSA configuration needs.
    """

    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        # Needed for 'rn' json detector parameters
        self.detector_readout_pattern = config['detector']['readout_pattern']
        config['detector']['max_total_groups'] = config['detector']['nint'] * config['detector']['ngroup']

        JWSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)

        slit = self.instrument.get('slit',None)
        if slit is not None:
            try:
                slit_config = self.slit_config[slit]
            except KeyError as e:
                msg = "Configuration for slit {} not specified".format(slit)
                raise DataError(value=msg)
            if self.mode == "msa":
                shutter_location = self.instrument['shutter_location']
                gap_config_file = slit_config.pop('gap')
                gap_config = read_json(os.path.join(self.ref_dir, gap_config_file), raise_except=True)
                self.shutter_locations = list(gap_config.keys())
                try:
                    self.gap = gap_config[shutter_location]
                except KeyError as e:
                    msg = "Shutter location not specified for MSA calculation: %s" % e
                    raise DataError(value=msg)
            else:
                self.gap = self.read_config_param(slit_config, "gap")
        self._loadpsfs()

    def get_wave_range(self):
        """
        Get the wavelength range of the current instrument configuration

        Returns
        -------
        range_dict: dict
            Contains the instrument wavelength range in microns described by:
                wmin - minimum wavelength
                wmax - maximum wavelength
        """
        disperser = self.instrument['disperser']
        aperture = self.instrument['aperture']
        filt = self.instrument['filter']
        # MSA shutter configuration is read in from a separate file in a different way

        if disperser is not None:
            # get the wavelength range from the gap configuration
            if filt in self.gap[disperser]:
                # g140m and g140h have different gap configs for each blocking filter
                g_wmin = self.gap[disperser][filt]["wave_min"]
                g_wmax = self.gap[disperser][filt]["wave_max"]
            else:
                g_wmin = self.gap[disperser]["wave_min"]
                g_wmax = self.gap[disperser]["wave_max"]

            # get the wavelength range from the configuration file
            c_wmin = self.range[aperture][filt]["wmin"]
            c_wmax = self.range[aperture][filt]["wmax"]

            # get the wavelength range over which the disperser efficiency is known
            wave_blaze = self.get_wave_blaze()
            d_wmin = wave_blaze.min()
            d_wmax = wave_blaze.max()

            # get the wavelength range over which the filter throughput is known
            wave_filter = self.get_wave_filter()
            f_wmin = wave_filter.min()
            f_wmax = wave_filter.max()

            # compare filter and disperser wavelength ranges
            if f_wmax < d_wmin or d_wmax < f_wmin:
                raise RangeError(value="Disperser and Filter wavelength ranges do not overlap.")
            wmin = max(f_wmin, d_wmin, c_wmin, g_wmin)
            wmax = min(f_wmax, d_wmax, c_wmax, g_wmax)
        else:
            wmin = self.range[aperture][filt]["wmin"]
            wmax = self.range[aperture][filt]["wmax"]

        range_dict = {'wmin': wmin, 'wmax': wmax}
        return range_dict

    def create_gap_mask(self, wave):
        """
        Use the gap configuration and a wavelength vector, wave, to build a masked array that
        masks out the location of the gap.  Wavelengths between and including both gap endpoints
        will be masked out.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to construct mask from

        Returns
        -------
        mask: numpy.ma 1D masked array
            1D array masked at the locations within the configured detector gap and 1.0 elsewhere
        """
        disperser = self.instrument['disperser']
        aperture = self.instrument['aperture']
        filt = self.instrument['filter']
        # MSA shutter configuration is read in from a separate file in a different way
        if hasattr(self, "gap"):
            gap = self.gap[disperser]

            if filt in gap:
                gap_start = gap[filt]['gap_start']
                gap_end = gap[filt]['gap_end']
            else:
                gap_start = gap['gap_start']
                gap_end = gap['gap_end']

            if gap_start is not None and gap_end is not None:
                masked_wave = ma.masked_inside(wave, gap_start, gap_end)
                mask = masked_wave / wave
                mask = ma.filled(mask, 0.0)
            else:
                mask = 1.0
        else:
            mask = 1.0
        return mask

    def get_internal_eff(self, wave, full_throughput=False):
        """
        Read in internal efficiency of NIRSpec. This is
        overloaded because the internal optical throughput is
        different for the NIRSpec IFU compared to MOS and Fixed Slit.

        This also multiplies the internal efficiency by the correction
        factor that showed up after commissioning.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Internal throughput as a function of wave
        """
        if self.mode in ["ifu", "ifu_ver"]:
            eff = self._get_throughput(wave, 'internal_ifu')
        elif self.mode in ["msa", "bots", "mos_conf", "mos_ver", "fixed_slit", "target_acq"]:
            eff = self._get_throughput(wave, 'internal_msa')
        else:
            msg = "Internal efficiency not configured for NIRSpec mode %s." % self.mode
            raise EngineInputError(value=msg)

        # correction factors for the filter+disperser combinations

        # generate a valid key only for the 3 modes that support correction factors
        # (this will change when more cf files are delivered)
        str_mode = "Nomode" # causes a fall thru to np.ones correction factor
        if "ifu" in self.mode:
            str_mode = "ifu"
        if "fixed_slit" in self.mode or "bots" in self.mode:
            str_mode = "fs"
        elif "mos" in self.mode:
            str_mode = "mos"

        disperser = self.instrument['disperser']
        if disperser is None:
            disperser = 'prism'
        filter = self.instrument['filter']
        if filter is None:
            filter = 'clear'

        key = "cf_%s_%s_%s" % (str_mode, filter, disperser)

        corr = self._get_throughput(wave, key, colname="obs/exp")

        return eff * corr

class NIRCam(JWSTInstrument):

    """
    Need to override __init__ to handle the complexity of the multiple coronagraphy apertures that we want to
    (optionally) hide from users.
    """

    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        if mode == "coronagraphy":
            # If we have one of the non-suffixed apertures, we must apply the appropriate detector suffix.
            # This code would turn mask210rsw into mask210rswlw if you use one suffixed aperture with the
            # other detector but that shouldn't work anyway. Nevertheless, let's put in an informative error
            # message.
            for detector in ["sw", "lw"]:
                if config["instrument"]["aperture"][-2:] == detector and config["instrument"]["detector"] != detector:
                    raise EngineInputError(f'Inconsistent selection of detector {config["instrument"]["detector"]} and aperture {config["instrument"]["aperture"]}')

            if config["instrument"]["detector"] != config["instrument"]["aperture"][-2:]:
                config["instrument"]["aperture"] = f"{config['instrument']['aperture']}{config['instrument']['detector']}"
        elif mode == "target_acq":
            # target_acq also uses coronagraphic apertures, but only has a 1:1 mapping.
            aperture_map = {"mask210r": "mask210rsw", "mask335r": "mask335rlw", "mask430r": "mask430rlw", "maskswb": "maskswbsw", "masklwb": "masklwblw", "sw": "sw", "lw": "lw",
                            "mask210rsw": "mask210rsw", "mask335rlw": "mask335rlw", "mask430rlw": "mask430rlw", "maskswbsw": "maskswbsw", "masklwblw": "masklwblw"}
            config["instrument"]["aperture"] = aperture_map[config["instrument"]["aperture"]]

        JWSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)

        self._loadpsfs()

    def _load_psf_library(self,psf_key):

        bar_offset=0
        if "masklwbsw" in self.get_aperture() or "maskswblw" in self.get_aperture():
            bar_offset = self.bar_offsets[self.instrument['paired_filter']]
        elif "masklwblw" in self.get_aperture() or "maskswbsw" in self.get_aperture():
            bar_offset = self.bar_offsets[self.instrument['filter'].split("_")[0]]

        psf_path = os.path.join(self.ref_dir, "psfs")
        library = PSFLibrary(self.get_wave_range(), path=psf_path, psf_key=psf_key, x_offset=bar_offset)

        return library

    def get_filter_eff(self, wave):
        """
        over-ride the filter efficiency because the narrow-band filters are in the pupil wheel,
        and therefore also go through a broad-band filter in the filter wheel (which doesn't have a clear).

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Filter throughput as a function of wave
        """

        if not hasattr(self, 'double_filters'):
            msg = "NIRCam requires a mapping that describes which filters are actually a combination of two filters."
            raise DataError(value=msg)

        if self.instrument['filter'] in self.double_filters:
            eff = self._get_throughput(wave, self.instrument['filter'])
            eff_pupil = self._get_throughput(wave, self.double_filters[self.instrument['filter']])
            eff *= eff_pupil
        else:
            eff = self._get_throughput(wave, self.instrument['filter'])
        return eff

    def get_internal_eff(self, wave, full_throughput=False):
        """
        Read in internal efficiency. For NIRCam there are separate internal efficiencies for the optics common
        to all modes, the throughput of the coronagrapher substrate, and the throughputs of the optical wedges
        that bring the coronagraphy elements into the field of view of the detectors.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Internal throughput as a function of wave
        """
        base_eff = self._get_throughput(wave, 'internal')
        coronagraphy_eff = 1.0
        wedge_eff = 1.0
        dichroic_eff = 1.0
        pupil_eff = 1.0

        # load the dichroic effects for the correct detector. Note that _get_throughput defaults to returning 1
        dichroic_eff = self._get_throughput(wave, 'dbs_{}'.format(self.instrument['detector']))

        # load the weak lens pupil transmission, if applicable
        if "wlp8" in self.instrument['aperture']:
            pupil_eff = self._get_throughput(wave, 'wlp8')

        if self.instrument['mode'] == 'coronagraphy':
            coronagraphy_eff = self._get_throughput(wave, 'coronagraphy_substrate')
            wedge_eff = self._get_throughput(wave, '{}_wedge_eff'.format(self.instrument['detector']))
        eff = base_eff * coronagraphy_eff * wedge_eff * dichroic_eff * pupil_eff

        return eff

    def dispersion_axis(self):
        """
        The dispersion axis is either along rows (the X axis) or along columns (Y axis).  By default
        it is along the X axis.  However, the GRISMC grating disperses along the Y axis as a way to
        help mitigate source crowding and confusion.

        Returns
        -------
        disp_axis: str
            Axis along with spectra are dispersed.  Allowed values are 'x' or 'y'.
        """
        if self.projection_type == 'slitless' and self.instrument['disperser'] == "grismc":
            disp_axis = "y"
        else:
            disp_axis = "x"
        return disp_axis

    def bar_width(self, x):
        """
        Width of MASKLWB or MASKSWB in arcsec as a function of X. The width at the center of the FOV is taken from the
        configuration as a function of what filter is being used.

        See NIRCam Coronagraph Operations Description, Version 4.1, Sept. 26, 2016, J.Stansberry, NIRCam Operations
        https://confluence.stsci.edu/download/attachments/52920601/NIRCam_CoronagraphOps_V4.1.pdf?version=1&modificationDate=1486495660042&api=v2
        This needs to be refactored to remove the hard-coded constants.

        Parameters
        ----------
        x: float
            X position (arcsec) in field of view

        Returns
        ------
        width: float
            Width of bar at X in arcsec
        """
        filt = self.instrument["filter"]
        if "paired_filter" in self.instrument:
            filt = self.instrument["paired_filter"]
        if filt not in self.bar_offsets:
            msg = "Invalid filter, %s, for MASKLWB/MASKSWB." % filt
            raise DataError(value=msg)
        center = self.bar_offsets[filt]
        if "maskswb" in self.instrument['aperture']:
            # the maskswb bar widens as x increases
            width = 0.2666 - 0.01777 * (center + x)
        elif "masklwb" in self.instrument['aperture']:
            # the masklwb bar is flipped: it narrows as x increases
            center = center * -1
            x = x * -1
            width = 0.5839 - 0.03893 * (center + x)
        else:
            msg = "bar_width() method only appropriate for MASKLWB and MASKSWB apertures."
            raise EngineInputError(value=msg)
        return width

    def get_detector_qe(self, wave):
        """
        Need to over-ride get_detector_qe() to handle the two different detectors. Which one to use is keyed off
        of the configured aperture.
        """
        try:
            detector = self.instrument['detector']
        except KeyError as e:
            msg = "NIRCam aperture configuration must include which detector the aperture belongs to, sw or lw. (%s)" % e
            raise DataError(value=msg)
        qe = self._get_throughput(wave, 'qe_{}'.format(detector))

        return qe


class NIRISS(JWSTInstrument):

    """
    Need to override _loadpsfs because the long wavelength filters go through a different mask (CLEARP) than
    the short-wavelength filters, and require a different set of PSFs that does overlap in wavelength.

    Also need to set up multiorder extraction masks for SOSS, and order-specific wavelength traces.

    Also need to override dispersion_axis() because the GR150C and GR700XD gratings disperse along the Y axis
    vs. the X axis like every other JWST disperser.
    """

    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        # these are required for calculation, but ok to live with config file defaults
        self.api_ignore = ['max_saturated_pixels', 'min_snr_threshold', 'aperture_size']

        JWSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)
        self._loadpsfs()

    def dispersion_axis(self):
        """
        The dispersion axis is either along rows (the X axis) or along columns (Y axis).  By default
        it is along the X axis.  However, the GR150C grating disperses along the Y axis as a way to
        help mitigate source crowding and confusion.

        Returns
        -------
        disp_axis: str
            Axis along with spectra are dispersed.  Allowed values are 'x' or 'y'.
        """
        if self.instrument['disperser'] in ("gr150c", "gr700xd"):
            disp_axis = "y"
        else:
            disp_axis = "x"
        return disp_axis

    def get_extraction_mask(self, order):
        """
        Each SOSS order has its own extraction mask. Use the specified order to build the key to look up the
        FITS file containing the mask.

        Parameters
        ----------
        order: int
            Order whose mask to read

        Returns
        -------
        mask: 2D np.ndarray
            Mask data
        """
        if order not in (1, 2, 3):
            msg = "SOSS order %d is not valid." % order
            raise EngineInputError(value=msg)

        key = "gr700xd_%d_mask" % order

        # substrip96 is a special case where there's only one possible mask
        if self.detector['subarray'] == 'substrip96':
            key += "96"

        path = self.paths.get(key, None)

        if path is None:
            msg = "No mask configured for SOSS order %d." % order
            raise DataError(value=msg)

        mask_file = os.path.join(self.ref_dir, path)
        try:
            mask = fits.getdata(mask_file)
        except Exception as e:
            msg = "Error reading mask data for SOSS order %d: %s" % (order, type(e))
            raise DataError(value=msg)

        return mask

    def _get_psf_key(self):
        """
        Short-wavelength filters need PSFs that have the CLEAR pupil mask (0.79-2.26 microns)
        Long-wavelength filters need PSFs that have the CLEARP mask (2.37-5.04 microns)
        (See https://jwst-docs.stsci.edu/display/JTI/NIRISS+Overview, Table 2)
        Because the ranges overlap when put on a grid, we need to switch between PSF libraries.
        """
        psf_key = self.instrument['aperture']
        if self.instrument['aperture'] == 'imager':
            if self.instrument['filter'] in self.lw_pupil:
                psf_key = "%s%s" % (self.instrument['aperture'], 'lw')
            else:
                psf_key = "%s%s" % (self.instrument['aperture'], 'sw')
        else:
            psf_key = self.instrument['aperture']

        return psf_key

    def get_trace(self, wave):
        """
        Read in spectral trace offsets from center of FOV. Currently wavelength-dependent spatial distortion is
        only required for SOSS mode. Other modes simply return 0's.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim trace data onto

        Returns
        -------
        trace: numpy.ndarray or float
            Spectral trace offset from center of FOV as a function of wave
        """
        if self.mode == "soss":
            disperser = self._get_disperser()
            key = "%s_wavepix" % disperser
            # handle the special case of substrip96
            if self.detector['subarray'] == 'substrip96':
                key += "96"

            # SOSS modes requires trace reference data to work properly. so raise exception if we can't load it.
            try:
                trace = self._interp_refdata(wave, key, colname='trace')
            except DataError as e:
                msg = "Spectral trace data missing for NIRISS SOSS, %s."
                raise DataError(value=msg)
        else:
            trace = np.zeros_like(wave)
        return trace

    def get_detector_pixels(self, wave):
        """
        Read in detector pixel positions for each wavelength in wave_pix

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim pixel position data onto

        Returns
        -------
        pixels: numpy.ndarray or None
            Detector pixel positions corresponding to each element of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            key = "%s_wavepix" % disperser
            # handle the special case of substrip96
            if self.detector['subarray'] == 'substrip96':
                key += "96"
            try:
                pixels = self._interp_refdata(wave, key, colname='detector_pixels', default=np.arange(len(wave)))
            except DataError as e:
                pixels = None
        else:
            pixels = None
        return pixels

class MIRI(JWSTInstrument):

    """
    The MIRI internal efficiency, detector readout patterns, etc. are more complex, and different than the other instruments,
    so some methods are redefined.
    """
    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        JWSTInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)
        self._loadpsfs()

    def _get_psf_key(self):
        """
        The MIRI coronagraphic target acquisition modes use a tiny box that does not include any of the obscuration, and
        only on filters that do not have the coronagraphic occulters included. Therefore, we're using different PSFs
        for them, with aperture names "fqpm1065ta", and similar.
        """
        psf_key = self.instrument['aperture']
        
        if (self.instrument['mode'] in ('target_acq')) and (self.instrument['aperture'] in ('fqpm1065', 'fqpm1140',
                                                                                            'fqpm1550', 'lyot2300')):
            psf_key = '{0:}ta'.format(self.instrument['aperture'])            

        return psf_key

    @property
    def qe_key(self):
        """
        MIRI has three different detectors with three different QE reference files.  Use this
        method to pick the right reference file key based on the configured aperture and overload self.qe_key.

        Returns
        -------
        key: string
            Key for looking up the appropriate QE reference file
        """
        key = "miri_{}_qe".format(self.instrument['detector'])

        return key

    def get_variance_fudge(self, wave):
        """
        In addition to a scalar fudge factor, MIRI also has a chromatic variance fudge that correlates with
        the quantum yield. The information posted in Issue #2167 suggests that they want the noise scaled by an extra
        factor of the quantum yield so that the SNR scales inversely with quantum yield. The MIRI team has been asked
        to provide the chromatic fudge factor they want as a separate reference file. Until that's delivered, we'll
        use the quantum yield squared to achieve the desired effect.
        """
        scalar_fudge = self.the_detector._get_variance_fudge(wave)
        q_yield, fano_factor = self.get_quantum_yield(wave)
        var_fudge = scalar_fudge * q_yield**2

        return var_fudge

    def get_internal_eff(self, wave, full_throughput=False):
        """
        Calculate MIRI internal efficiency which is rather more complicated than the other instruments

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate efficiency onto

        Returns
        -------
        eff: numpy.ndarray or float
            Internal efficiency as a function of wave
        """
        aperture = self.instrument['aperture']
        mirror_eff = self.mirror_eff
        mirror_cont = self.mirror_cont
        n_refl = self.n_reflections[aperture]
        refl_eff = mirror_eff ** n_refl
        internal_eff = refl_eff

        # mirror contamination factor
        internal_eff = internal_eff * mirror_cont

        return internal_eff

    def get_disperser_eff(self, wave):
        """
        Overloaded here because disperser efficiency is keyed off of the aperture rather than disperser

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        disperser_eff: numpy.ndarray or float
            Disperser efficiency as a function of wave
        """
        if self.instrument['disperser'] is not None:
            key = "{}_{}".format(self.instrument['aperture'],self.instrument['disperser'])
            disperser_eff = self._get_throughput(wave, key)
        else:
            disperser_eff = 1.
        return disperser_eff

    def _get_dispersion_key(self):
        """
        Overload this because the key is constructed from both the aperture and disperser rather
        than the disperser alone.

        Returns
        -------
        key: str
            Key used to get dispersion file out of self.paths
        """
        disperser = self.instrument['disperser']
        aperture = self.instrument['aperture']
        key = "%s_%s_disp" % (aperture, disperser)
        return key

    def dispersion_axis(self):
        """
        The dispersion axis is either along rows (the X axis) or along columns (Y axis).  By default
        it is along the X axis.  However, the P750L grating disperses along the Y axis

        Returns
        -------
        disp_axis: str
            Axis along with spectra are dispersed.  Allowed values are 'x' or 'y'.
        """
        if self.instrument['disperser'] == "p750l":
            disp_axis = "y"
        else:
            disp_axis = "x"
        return disp_axis

def name_mapper(name=None):
    """
    General Purpose name remapping function
    If not fed a name, it returns the mapping dictionary
    If fed a name, it returns either the mapped name or the name (if no mapping is defined)
    Parameters
    ----------
    name: string
        The name of a JWST object
    Returns
    -------
    dictionary or string
        Either the remapped string (where necessary) or the complete mapping dictionary.
    """
    short_str_mappings = {
        'nircam ssgrism':  'nircam lw_tsgrism',
        'nirspec msa':     'nirspec mos',
        'ssgrism':         'lw_tsgrism',
        'msa':             'mos',
    }
    if name is None:
        return short_str_mappings
    else:
        if name in short_str_mappings:
            return short_str_mappings[name]
        else:
            return name



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pickle
import os
import datetime
import traceback
from functools import reduce, wraps
import time
import sys

TIMING=True

class AutoVivification(dict):
    """
    Implementation of perl's autovivification feature.
    https://stackoverflow.com/questions/651794/whats-the-best-way-to-initialize-a-dict-of-dicts-in-python
    """
    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value

def init(config):
    #
    # This is a reasonably standard method to have a global variable and then accessible via other objects.
    #
    # https://stackoverflow.com/questions/13034496/using-global-variables-between-files
    #
    global debugarrays
    debugarrays = DebugArrays(config)

class DebugArrays():
    """
    The class implements the functionality to turn off and on storage/reporting of intermediate
    data computed in the engine code.

    The "debugarrays" can be put at the bottom of the jeng file and would look like:

            ....
                "strategy": {
                "aperture_size": 0.53,
                "axis": "x",
                "background_subtraction": true,
                "display_string": "Target Acquisition Centroid",
                "method": "tacentroid",
                "target_source": 1,
                "target_type": "",
                "target_xy": [
                    0.0,
                    0.0
                ],
                "units": "arcsec"
            },
            "debugarrays": {
                "etc3D": {
                    "ipc_convolve":  {
                        "type": "report",
                        "keys": ["etc3D", "ipc_convolve"],
                        "description": "Testing this to include this info"
                    }
                },
                "astro_spectrum": {
                    "__init__": {
                        "type": "report",
                        "keys": ["astro_spectrum", "init"]
                    },
                    "init_file":
                        "type": "pickle",
                        "filename": "/tmp/bob.pck",
                        "description": "We want to store this info as it is important"
                }
            }


    Then in the astro_spectrum file is, for example:

                debug_utils.debugarrays.store('astro_spectrum', '__init__',
                                      {
                                          'model_scene': model_scene.int[:, :, windex],
                                          'kernel': kernel,
                                          'intensity': self.intensity,
                                          'description': 'Not a single point source near or at the center.'
                                      })

                debug_utils.debugarrays.store('astro_spectrum', 'init_file',
                                      {
                                          'model_scene': model_scene.int[:, :, windex],
                                          'kernel': kernel,
                                          'intensity': self.intensity,
                                          'description': 'Not a single point source near or at the center.'
                                      })


    """

    def __init__(self, config):
        """
        Called with the config data from the jeng file.

        :param config:
        """

        self.config = config
        self.report = AutoVivification()

    def store(self, filename, label, data):
        """
        This function should be called from within the code in order to save a set of data.

        There are currently two types that are enabled:
            1. report - the data passed to the function will be stored here, internally, and then the
                        get_report function needs to be called from within a function in report.py

            2. pickle - this will immediately save the data to a pickle file specified in teh filename.

        :param filename:
        :param label:
        :param data:
        :return:
        """

        # Add in the stack for more information
        data['stack'] = traceback.extract_stack()

        # First, let's see if, and then how, we are going to store the data
        if self.config.get(filename) and self.config.get(filename).get(label):
            dict_entry = self.config.get(filename).get(label)

            # Temporarily store into a variable here which will then need to be accessed by calling get_report()
            if dict_entry.get('type') == 'report':
                self._nested_set(self.report, dict_entry.get('keys'), data, extra_description=dict_entry.get('description'))

            # Immediately store the data into a pickle file.
            elif dict_entry.get('type') == 'pickle':

                # Check to see if the clobber flag is set... if not then we will have to check
                # to see if the file exists, if it does then append a date or something
                # TODO: Do some checking on the filename, directory existance etc
                filename = dict_entry.get('filename')
                if not dict_entry.get('clobber') and filename and os.path.isfile(filename):
                    tfile_name, t_extension = os.path.splitext(filename)
                    thedate = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
                    filename = tfile_name + thedate + t_extension

                if not filename:
                    Exception('No filename')

                if not dict_entry.get('clobber') and os.path.isfile(filename):
                    Exception('No clobber but file exists')

                self._save_pickle(filename, data, extra_description=dict_entry.get('description'))

            else:
                Exception('debug_utils, unknown type {}'.format(dict_entry.get('type')))

    def get_report(self):
        """
        Theoretically this will be called from report.py so the internal report dict will be passed out (and
        assumably added to the reporting dictionary in report.py

        :return:  self.report - dictionary of output
        """

        return self.report

    def _save_pickle(self, filename, data, extra_description=None):
        """
        It could be that larger datasets might need to be saved in a file. Save the passed data to a pickle file.

        :param filename: Output Pickle file.
        :param data: The data to output.
        :return:
        """

        # Add in the extra description information from the jeng file
        if extra_description:
            data['extra_description'] = extra_description

        fp = open(filename, 'wb')
        pickle.dump(data, fp)
        fp.close()

    def _nested_set(self, dic, keys, value, extra_description=None):
        """
        The list of hierarchical keys will be added to the dict and the value will be added at that level. For example:

            keys = ['first', 'second', 'third']
            value = 5

        then
            dic['first']['second']['third'] = 5

        Based on:
        https://stackoverflow.com/questions/13687924/setting-a-value-in-a-nested-python-dictionary-given-a-list-of-indices-and-value

        :param keys:
        :param value:
        :return:
        """

        # Add in the extra description information from the jeng file if we can
        if extra_description and type(value) == dict:
            value['extra_description'] = extra_description

        for key in keys[:-1]:
            dic = dic.setdefault(key, {})
        dic[keys[-1]] = value

# Note that this is Py3.3+-specific syntax
def timethis(func):
    """
    A Decorator that reports (or logs) execution time
    """
    @wraps(func) # to avoid losing the calling signature and name
    def wrapper(*args, **kwargs):
        if TIMING: # set this in the header of a file to activate timing in its functions
            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()
            # get the current stack
            stack = traceback.extract_stack()
            msg = "{:0.8f} Timethis {} ".format(end-start, func.__qualname__)
            if len(stack) > 2:
                # select the last 3 (or fewer) entries for the message
                # last entry is always this wrapper itself
                for x in range(-2, -1*(min(len(stack),5)),-1):
                    # each item in the stack is a FrameSummary. Convert to string,
                    # slice at " in ", take what's after, and remove the trailing >
                    msg += "-- {} ".format(str(stack[x]).split(" in ")[-1][:-1])
            print(msg)
            return result
        else:
            return func(*args, **kwargs)
    return wrapper

# https://goshippo.com/blog/measure-real-size-any-python-object/
def get_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from astropy import units as u
import synphot as syn


# These are the wavelength and flux units that are used internally for all calculations.
PANDEIA_WAVEUNITS = u.micron
PANDEIA_FLUXUNITS = u.mJy

UNIT_MAP = {"flam": syn.units.FLAM, "fnu": syn.units.FNU, "abmag": u.ABmag, 
            "vegamag": syn.units.VEGAMAG, 'njy': u.nJy, "ujy": u.uJy, "mjy": u.mJy,
            "jy": u.Jy, "MJy": u.MJy, "microns": u.micron, "micron": u.micron, 
            "angstrom": u.AA, "nm": u.nm, "um": u.um, "mm": u.mm, "m": u.m}

# Absolute and relative tolerance values to use to check if values are close enough to be considered equal.
# See https://docs.scipy.org/doc/numpy/reference/generated/numpy.isclose.html for details on how they're applied
# in comparisons.
PANDEIA_ATOL = 2.0e-8
PANDEIA_RTOL = 1.0e-6

# A necessary conversion constant for converting mJy to photons/s. Pedigree is unknown, value originates in this commit:
# https://github.com/spacetelescope/scamp/commit/025c4098e2bcad6ad7ac65368d877924c374bd69
CONVERSION_CONSTANT = 1.5091905

# Pandeia subsamples lines in order to accurately reproduce thin spectral lines in a spectrum. This subsampling must
# extend over a large enough range to contain most of the wings and accurately reproduce the line.
#
# spectral_line_window determines the size of the wavelength window over which lines are subsampled. Values above 15
#  (decided empirically, see Issue #3292) are necessary to contain the wings of the line.
# spectral_line_subsample determines how much the spectrum is subsampled within the window to reproduce the shape of
#  the profile.
SPECTRAL_LINE_WINDOW = 20
SPECTRAL_LINE_SUBSAMPLE = 3

# In practice, a value of 200 samples within an imaging configuration's wavelength range (i.e. filter bandpass) should
# be more than enough to produce an accurate image. Note that because we use synphot to resample, even the flux of
# narrow lines is conserved.
SPECTRAL_MAX_SAMPLES = 200

# Pandeia's spectrum processing can use a considerable amount of memory for large fields (like SOSS). To handle this,
# we divide the spectra into chunks of wavelength planes (if necessary) at the cost of a few seconds of run time. There
# may be a more elegant way to handle this...
SPECTRAL_MAX_CHUNK = 1000

# Pandeia's FOV size (for the spatial axes of the scene cubes) is the largest of either the scene size, the PSF size,
# or the configured instrument's default FOV size. We need to add a small buffer to the scene size to adequately contain
# the full scene after sources have been represented by PSFs
SCENE_FOV_PIXBUFFER = 20

# The ConvolvedSceneCube function uses astropy.convolve_fft to produce a convolved scene cube. This cube cannot be
# created if any of the pixels are too bright - even getting close to the maximum floating point number that can be
# stored in a numpy array. Empirical studies show that values greater than 1e33 still cause astropy.convolve_fft to fail
# We have set this to 1e32 to be safe.
SCENE_MAX_PIXELVAL = 1e32

# The profile generator needs to find pixels whose locations are 0, but due to floating point math, the values won't
# necessarily be exactly zero. Set a tiny value accordingly.
POS_EPSILON = 1e-8

# The profile generator needs to oversample the central pixel of high-sersic-index sersic profiles. They can be
# incredibly peaky, such that the amount of flux would be drastically overstated if we were to simpy set the central
# pixel to the value at the exact center. Instead, we need to oversample that pixel and compute a more accurate flux for
# the central pixel of the sersic. The process by which the number 101 was arrived at can be found here:
# Code PR #3892#issuecomment-397713379
SERSIC_OVERSAMPLE = 101

# The default dispersion is meant for use when attempting to create backgrounds for order 0 of slitless and multiorder
# modes, because order 0 itself is not dispersed but our algorithms demand it.
DEFAULT_DISPERSION = 0.01

# Minimum to clip values to, for avoidance of divsion-by-zero errors in exposure timing, profile generation, etc.
MIN_CLIP = 1e-10

# How many parts to split the extraction aperture into (for the TASpec strategy)
EXTRACTION_NSPLIT=1000



#!/usr/bin/env python
#
# This is a self-test of the pandeia-repo BIT translation code.
#
# Assume:  the pyetc test repo contains true/correct translations
#          from apt queries to peng files under: pyetc/test/engine/bit
#          covering all nightly tests of these queries.  This may not
#          be all the queries but it (I believe) is the full unique set.
#
# Assume:  each peng file in the pyetc/test/engine/bit tree includes the
#          apt query used to generate it as a comment at the top
#
# The Plan:  find all of these files and for every one, read the apt
#            query, use pandeia repo code to generate the peng, verify
#            the the output peng is the same as what was in pyetc's repo
#
# How to use: Change the repo_dir variable in main function to the pyetc
#             repo in your local machine. 
#             Run with `python bit_test_query_to_peng.py`.

import os, sys
from pandeia.engine.helpers.bit import bit
from pandeia.engine.helpers.bit.pyetc_util import write_dict, fast_write_dict

from pandeia_test import utils


def create_peng_ref_obj(fname):
    """ read in a single peng and get all of its meta info to make a dict """
    retval = {}
    f = open(fname, 'r')
    retval['buf'] = f.read().strip() # no need for final newline
    f.close()

    pathexp = 'pyetc/test/engine/bit/'
    idx = fname.index(pathexp) # let throw on err
    retval['fname'] = fname # e.g. '/Users/bob/dev/pyetc/test/engine/bit/acs/imag/sbc/sbc/f115lp/f5v_mag15v.peng'
    retval['testname'] = fname[idx+len(pathexp):] # e.g. 'acs/imag/sbc/sbc/f115lp/f5v_mag15v.peng'

    lines = retval['buf'].split('\n')
    assert len(lines) > 5, 'Unexpected peng format for: '+fname
    # header is like:
    # 'url query: type=bot&magnitude=15&config=ACS/SBC&filter0=F115LP&aperture=SBC&requestCount=2&request0=BrightestPixelRate&request1=ImageRate&stellarType=F5V&magType, test name: acs/imag/sbc/sbc/f115lp/f5v_mag15v.peng'
    header = lines[0]
    assert header[0] == '#', 'Unexpected 1st line of peng: '+lines[0]
    parts = header.split(',')
    query = parts[0].split(' ')[-1]
    retval['query'] = query

    # final sanity check
    tname = parts[-1].split(' ')[-1]
    tname = tname.strip("'")
    assert tname == retval['testname'], f'test name from fname: {retval["testname"]} != test name from header: {tname}'

    return retval


def find_all_pyetc_bit_pengs(pyetcrepo_test_engine_bit_dir):
    """ Return a list of dict-objects, one for each apt query found at the top comment of
    a peng file in the pyetc/test/engine/bit repo. """

    # check input
    for inst in ('acs','cos','stis','wfc3ir','wfc3uvis'):
        assert os.path.exists(pyetcrepo_test_engine_bit_dir+'/'+inst), 'Cannot find: '+pyetcrepo_test_engine_bit_dir+'/'+inst

    # all engine/bit pengs from pyetc/test
    pengs = utils.rglob(pyetcrepo_test_engine_bit_dir,'*.peng')
    retval = [create_peng_ref_obj(p) for p in pengs]
    return retval


def main():

    # First find all of our "truth" files.
    #
    # list[dict{fname: str, query: str, buf: str}]
    repo_dir = '/Users/sontag/dev/pyetc/test/engine/bit'
    bit_peng_files = find_all_pyetc_bit_pengs(repo_dir)

    # loop over them
    for peng_obj in bit_peng_files:

        print('')
        # use pandeia code to convert to peng
        orig_file = peng_obj['fname']
        query = peng_obj['query']
        testname = peng_obj['testname']
        pweb_dict, peng_dict = bit.query_to_peng(query)

        # pweb currently gets dropped on floor
        new_file = orig_file+'_regen'

        # write new file
        if os.path.exists(new_file):
            os.remove(new_file)
        fast_write_dict(peng_dict, new_file, comment=f"'url query: {query}, test name: {testname}'")
        print('Orig: ', orig_file)
        print('Wrote:', new_file)

        # diff it for pass/fail but don't capture (for now)
        if sys.platform=='darwin':
            diffcmd = "/opt/sw/bin/diff"
        else:
            diffcmd = "/usr/bin/diff"
        assert os.path.exists(diffcmd), 'Cannot find: '+diffcmd
        # !!! expect MJD diffs for now !!!
        diffcmd += " --ignore-matching-lines='obsmode.*,MJD.59' "+orig_file+" "+new_file

        x = os.system(diffcmd)
        assert x == 0, 'New file is different: '+new_file
        os.remove(new_file)

    print(f'\nJust looped over {len(bit_peng_files)} in {repo_dir}')
    print('All tests passed')


#
# main routine
#
if __name__=='__main__': # in case something else imports this file

    main()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .custom_exceptions import EngineInputError
from .utils import recursive_subclasses, merge_data

from .detector import Detector, Detector_MultiAccum, Detector_UnevenMultiAccum, Detector_SingleAccum
from .detector import H2RG, H4RG, SiAs
from .detector import CCD, MAMA, H1R, XDL

def DetectorFactory(webapp=False, config={}, **kwargs):
    """
    Function to take a detector type and configuration data and build/return an appropriately
    configured instance of the desired Detector class.

    Parameters
    ----------
    webapp: bool
        Switch to toggle strict API checking
    config: dict
        Configuration data segments needed to set up the exposure calculation
    **kwargs: keyword/value pairs
        Additional configuration data
    """

    types = recursive_subclasses(Detector)
    detectors = [t.__name__.lower() for t in types]
    type_map = dict(list(zip(detectors, types)))

    exposure_string = config["detector_config"]["det_type"].lower()

    cls = type_map[exposure_string](webapp=webapp, config=config, **kwargs)
    return cls



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import scipy.special as sp
import astropy.units as u
from scipy.integrate import dblquad

from .config import DefaultConfig
from .coords import Grid
from .constants import POS_EPSILON, SERSIC_OVERSAMPLE, MIN_CLIP
from .utils import recursive_subclasses
from .custom_exceptions import EngineInputError


def ProfileFactory(psf_upsamp, config={}, **kwargs):

    """
    Function to take profile configuration data and build/return a configured instance of
    the appropriate distribution generator.

    Parameters
    ----------
    webapp: bool
        Switch to toggle strict API checking
    config: dict
        Configuration data in engine API dict format
    **kwargs: keyword/value pairs
        Additional configuration data
    """
    #all_config = merge_data(config, dict(**kwargs))
    all_config = config
    if 'geometry' in all_config.shape:
        method = all_config.shape['geometry'].replace('_', '')
    else:
        msg = "Must specify type of profile to create."
        raise EngineInputError(value=msg)

    # we need to run the API checks here after merging in the normalization type-specific defaults
    # if webapp:
    #    self._api_checks(all_config)

    types = recursive_subclasses(Distribution)
    methods = [t.__name__.lower().replace('distribution', '') for t in types]
    type_map = dict(list(zip(methods, types)))

    if method not in methods:
        msg = "Unsupported or not yet implemented profile type: %s" % method
        raise EngineInputError(value=msg)
    else:
        cls = type_map[method](config, psf_upsamp, **kwargs)
        return cls


class Distribution:
    """
    Abstract class for source profiles.
    Normally, when defining undefined abstract properties, we use NotImplementedError, with an eye toward implementing
    that feature later. However, in this case, when the normalization methods are not implemented it is because there
    is no robust sensible definition for them, and we will NEVER implement them. Therefore, the abstract classes
    produce errors to that effect.

    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.
    psf_upsamp: int
        How oversampled the source's grid is.

    Attributes
    ----------
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    pa: float
        Position angle of major axis
    xrot: np.ndarray
        rotated and shifted indicies, X axis values
    yrot: np.ndarray
        rotated and shifted indicies, Y axis values
    pix_area_sqarcsec: float
        area of a detector pixel in square arcseconds
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    Methods
    -------
    normalized()
        return the normalized 2D profile
    raw()
        return the raw 2D profile

    """
    def __init__(self, src, psf_upsamp):
        self.pa=src.position['orientation']
        self.xoff=src.position['x_offset']
        self.yoff=src.position['y_offset']
        self.grid = src.grid
        yrot, xrot = self.grid.shift_rotate(self.yoff, self.xoff, self.pa)
        self.xrot = xrot
        self.yrot = yrot
        self.pix_area_sqarcsec = src.grid.xsamp * src.grid.ysamp
        self.norm_method = src.shape['norm_method']
        self.geometry = src.shape['geometry']
        self.psf_upsamp = psf_upsamp

    def pixelscale(self):
        if self.surf_area_units in ['sr']:
            arcsec2 = u.arcsec * u.arcsec
            normfactor = self.pix_area_sqarcsec / u.sr.to(arcsec2)  # convert area in steradians to area in pixels
        elif self.surf_area_units in ['arcsec^2', None]: # 'None' should be an option because integrated flux
                                                        # shouldn't have units (internally, the grid is arcsec)
            normfactor = self.pix_area_sqarcsec
        else:
            msg = "Unsupported surface area unit: %s" % self.surf_area_units
            raise EngineInputError(value=msg)

        return normfactor
    def normalize(self):
        options = {'integ_infinity': self.integrate_infinity, 'surf_scale': self.surface_scale,
                       'surf_center': self.surface_center}
        options[self.norm_method]()

    def integrate_infinity(self):
        msg = "Normalization method {0:} not supported for profile type {1:}".format(self.norm_method, self.geometry)
        raise EngineInputError(value=msg)

    def surface_scale(self):
        msg = "Normalization method {0:} not supported for profile type {1:}".format(self.norm_method, self.geometry)
        raise EngineInputError(value=msg)

    def surface_center(self):
        msg = "Normalization method {0:} not supported for profile type {1:}".format(self.norm_method, self.geometry)
        raise EngineInputError(value=msg)

    def normalized(self):
        return self.prof * self.norm_val

    def raw(self):
        return self.prof
    
    def brightest(self):
        # make a new 1-pixel grid, shift the source back to center
        self.xoff = 0
        self.yoff = 0
        self.grid = Grid(self.grid.xsamp, self.grid.ysamp, self.psf_upsamp, self.psf_upsamp)
        self.generate()
        norm_prof = self.normalized()
        return np.sum(norm_prof), self.extended


class PointDistribution(Distribution):
    """
    Use pandeia.engine.coords.Grid.point_source() to generate a point source with
    subpixel positioning.

    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    """
    def __init__(self, src, psf_upsamp):
        self.xoff = src.position['x_offset']
        self.yoff = src.position['y_offset']
        self.extended = False
        self.base_center = 1.0
        self.psf_upsamp = psf_upsamp
        self.grid = src.grid

        self.generate()

    def generate(self):
        # This function both creates and fills the point source profile. It should
        # eventually be refactored to work the same way as other sources
        self.prof = self.grid.point_source(xoff=self.xoff, yoff=self.yoff)

        self.norm_val = 1
        self.extended = False

class SersicDistribution(Distribution):
    """
    Create a 2-dimensional elliptical source on the current grid. The intensity
    profile is described by a Sersic profile, I(r) = I(0) * exp(-b*((r/r_scale)**(1/n))-1),
    where r_scale is the effective radius such that I(r) = 0.5*I(0) and n is the Sersic index.
    The ellipticity is governed by specifying major and minor axis scale lengths
    separately.


    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    minor: float
        Minor axis scale length
    major: float
        Major axis scale length
    pa: float
        Position angle in degrees of major axis measured positive in +X direction
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    sersic_index: float
        Sersic profile shape parameter. 0.5 => gaussian, 1.0 => exponential, 4.0 => de Vaucouleurs
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    Methods
    -------
    _integrate_infinity():
        We use an analytical function to represent the integration of the sersic function
        from -Inf to +Inf in both axes. This will account for flux that falls
        outside of the FOV.
    _surface_center():
        Normalization is to the surface brightness of the center of the profile. This
        formulation is already normalized at that point.
    _surface_scale():
        Normalization is to the surface brightness of a point at the e-folding radius of
        this profile, where intensity is 1/e of the central intensity.
    sersic_generator():
        Handle subpixel-positioning and proper modeling of the central pixel (which can be too
        bright when sersic_index > 3)
    """
    def __init__(self, src, psf_upsamp):
        Distribution.__init__(self, src, psf_upsamp)
        self.norm_method = src.shape['norm_method']
        self.major=src.shape['major']
        self.minor=src.shape['minor']
        self.sersic_index=src.shape['sersic_index']
        self.surf_area_units = src.shape['surf_area_units']
        # the actual value of b. Formula taken from astropy's sersic2d shape.
        self.b = sp.gammaincinv(2*self.sersic_index,0.5)
        self.extended = True
        self.base_center = np.e**(self.b)

        self.generate()

    def generate(self):
        self.sersic_generator()

        self.normalize()

    def sersic_generator(self):
        '''
        Generic function to create a properly normalized sersic, taking into account:
         * Subpixel positioning (off-screen and on) by replicating some of the grid.point_source functionality
         * Realistic peak counts, by creating a highly oversampled grid covering just the central pixel

        NB: Subclasses should override sersic_func to specify the appropriate equation for the desired distribution.
        This method will rely on the subclass's sersic_func method.

        '''
        # oversample the central pixel to compute an actual flux for what may be an extremely tiny peak
        center_grid = Grid(self.grid.xsamp/(SERSIC_OVERSAMPLE-1), self.grid.ysamp/(SERSIC_OVERSAMPLE-1),
                           SERSIC_OVERSAMPLE, SERSIC_OVERSAMPLE)
        yrot,xrot = center_grid.shift_rotate(0, 0, self.pa)
        center_pixel = self.sersic_func(yrot,xrot,self.major,self.minor,self.sersic_index)
        # this should only correct for the oversampling compared to the usual grid
        self.center_flux = np.sum(center_pixel) / (SERSIC_OVERSAMPLE**2)

        # self.prof is a container to hold the actual profile
        self.prof = np.zeros_like(self.grid.x)
        # find the pixel-snapped locations of the lower left of the four pixels, and how far off the actual center it is
        # note that this doesn't actually have to be on the detector grid, this is just to properly split up the flux.
        lx = (self.xoff // self.grid.xsamp) * self.grid.xsamp
        rx = (self.xoff % self.grid.xsamp) * self.grid.xsamp
        ly = (self.yoff // self.grid.ysamp) * self.grid.ysamp
        ry = (self.yoff % self.grid.ysamp) * self.grid.ysamp

        # make a list of pixel positions and the amount of flux distributed to each.
        pixels = [[lx, ly, (1 - rx) * (1 - ry)], [lx, ly + self.grid.ysamp, (1 - rx) * ry],
                      [lx + self.grid.xsamp, ly, rx * (1 - ry)],
                      [lx + self.grid.xsamp, ly + self.grid.ysamp, rx * ry]]
        # create profiles centered in each of the four pixels, and scale them to the amount of flux
        for tx,ty,frac in pixels:
            # skip pixel if it adds no flux. In theory, the flux will be split between 1, 2, or 4 pixels.
            # if it's exactly centered on a pixel (say, at 0,0) the flux will be entirely on one pixel.
            # if it's offset slightly in X or Y, it will be split across two pixels
            # if it's offset slightly in both X and Y, it'll be split among four pixels.
            if frac != 0:
                # xrot and yrot contain the coordinates of each detector pixel center, in the coordinate system of the
                # offset and centered profile.
                yrot, xrot = self.grid.shift_rotate(ty, tx, self.pa)
                partial = self.sersic_func(yrot, xrot, self.major, self.minor, self.sersic_index)
                # the zero-value may not be zero, but something exceedingly small
                yzero = np.min(np.abs(yrot))
                xzero = np.min(np.abs(xrot))
                # if the central pixel is actually on the detector grid (and the sersic_index is greater than 0.5), we
                # need to find it and replace its flux with the corrected flux for the central pixel
                if np.abs(xzero) < POS_EPSILON * self.grid.xsamp and np.abs(yzero) < POS_EPSILON * self.grid.ysamp:
                    xvals = np.where(np.abs(xrot.flatten()) < POS_EPSILON * self.grid.xsamp)[0]
                    yvals = np.where(np.abs(yrot.flatten()) < POS_EPSILON * self.grid.ysamp)[0]
                    xy = np.intersect1d(xvals,yvals)
                    xval = xy // self.grid.shape[1]
                    yval = xy % self.grid.shape[1]
                    partial[xval, yval] = self.center_flux
                self.prof += partial * frac

    def integrate_infinity(self):
        # integrate the Sersic profile to get the total flux for normalization, including flux outside the FOV
        # http://ned.ipac.caltech.edu/level5/March05/Graham/Graham2.html
        integral = self.major * self.minor * 2 * np.pi * self.sersic_index * np.exp(self.b)/(self.b**(2*self.sersic_index))* sp.gamma(2 * self.sersic_index)
        self.norm_val = self.pixelscale() / integral

    def surface_scale(self):
        # for the base sersic profile, the flux at R=Re is 1. e^(-b* (1-1))
        self.norm_val = self.pixelscale()

    def surface_center(self):
        # for the base sersic profile, the flux at R=0 is e**b, so we must multiply by e**-b
        self.norm_val = self.pixelscale() * np.e**(-1*self.b)

    def sersic_func(self, y, x, major, minor, index):
        """
        Implement the Sersic intensity profile as a function to actually fill the grid with the profile.
        This is Equation 1 of Graham & Driver (2005) 2005PASA...22..118G

        This distribution function is also replicated in client/js/scenepage.js and should be kept in sync with any
        changes made here.

        Parameters
        ----------
        y: float or numpy.ndarray
            Y values for evaluating function
        x: float or numpy.ndarray
            X values for evaluating function
        major: float
            Major axis scale length
        minor: float
            Minor axis scale length
        index: float
            Sersic index

        Returns
        -------
        profile: float or numpy.ndarray
            Float or array containing evaluated Sersic profile
        """
        dist = np.sqrt((x / major)**2.0 + (y / minor)**2.0)
        # This is Equation 1 of Graham & Driver (2005) 2005PASA...22..118G
        profile = np.exp( -self.b * (dist**(1.0 / index) - 1) )

        return profile


class SersicScaleDistribution(SersicDistribution):
    """
    Create a 2-dimensional elliptical source on the current grid. The intensity
    profile is described by a Sersic profile, I(r) = I(0) * exp(-(r/r_scale)**(1/n)),
    where r_scale is the scale length where I(r) = I(0)/e and n is the Sersic index.
    The ellipticity is governed by specifying major and minor axis scale lengths
    separately.
    This is Equation 14 of Graham & Driver (2005) 2005PASA...22..118G


    This was the original implementation of the Sersic profile in pandeia, up through version 1.2.2. However, based on
    community feedback, it was replaced by the above SersicDistribution class. This class was retained but renamed to
    permit backwards compatibility in the web app.


    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    minor: float
        Minor axis scale length
    major: float
        Major axis scale length
    pa: float
        Position angle in degrees of major axis measured positive in +X direction
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    sersic_index: float
        Sersic profile shape parameter. 0.5 => gaussian, 1.0 => exponential, 4.0 => de Vaucouleurs
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    Methods
    -------
    _integrate_infinity():
        We use an analytical function to represent the integration of the sersic function
        from -Inf to +Inf in both axes. This will account for flux that falls
        outside of the FOV.
    _surface_center():
        Normalization is to the surface brightness of the center of the profile. This
        formulation is already normalized at that point.
    _surface_scale():
        Normalization is to the surface brightness of a point at the e-folding radius of
        this profile, where intensity is 1/e of the central intensity.
    """

    def __init__(self, src, psf_upsamp):
        Distribution.__init__(self, src, psf_upsamp)
        self.norm_method = src.shape['norm_method']
        self.major = src.shape['major']
        self.minor = src.shape['minor']
        self.sersic_index = src.shape['sersic_index']
        self.surf_area_units = src.shape['surf_area_units']
        self.extended = True
        self.base_center = 1.0

        self.generate()

    def generate(self):

        # The sersic_generator is inherited from the parent SersicDistribution class, but will correctly use this
        # class's own sersic_func(), which contains the appropriate equation.
        self.sersic_generator()

        self.normalize()

    def integrate_infinity(self):
        # integrate the Sersic profile to get the total flux for normalization, including flux outside the FOV
        # http://ned.ipac.caltech.edu/level5/March05/Graham/Graham2.html
        integral = self.major * self.minor * 2 * np.pi * self.sersic_index * sp.gamma(2*self.sersic_index)
        self.norm_val = self.pixelscale() / integral

    def surface_scale(self):
        self.norm_val = np.e * self.pixelscale()

    def surface_center(self):
        self.norm_val = self.pixelscale()

    def sersic_func(self, y, x, major, minor, index):
        """
        Implement the Sersic_scale intensity profile as a function to actually fill the grid with the profile.

        This distribution function is also replicated in client/js/scenepage.js and should be kept in sync with any
        changes made here.

        Parameters
        ----------
        y: float or numpy.ndarray
            Y values for evaluating function
        x: float or numpy.ndarray
            X values for evaluating function
        major: float
            Major axis scale length
        minor: float
            Minor axis scale length
        index: float
            Sersic index (traditionally, a number between 0.5 (gaussian) and 10)

        Returns
        -------
        profile: float or numpy.ndarray
            Float or array containing evaluated Sersic profile
        """
        dist = np.sqrt((x / major) ** 2.0 + (y / minor) ** 2.0)
        # This is Equation 14 of Graham & Driver (2005) 2005PASA...22..118G
        profile = np.exp(-dist ** (1.0 / index))
        return profile


class Gaussian2dDistribution(SersicScaleDistribution):
    """
    The Gaussian 2D profile is a special case of the Sersic_scale profile, where sersic index = 0.5.

    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    minor: float
        Minor axis scale length
    major: float
        Major axis scale length
    pa: float
        Position angle in degrees of major axis measured positive in +X direction
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    """
    def __init__(self, src, psf_upsamp):
        Distribution.__init__(self, src, psf_upsamp)
        self.norm_method = src.shape['norm_method']
        self.major=src.shape['major']*np.sqrt(2.0) # to match the usual definition of a gaussian
        self.minor=src.shape['minor']*np.sqrt(2.0)
        self.sersic_index=0.5
        self.surf_area_units = src.shape['surf_area_units']
        self.extended = True
        self.base_center = 1.0

        self.generate()

    def generate(self):
        # The sersic_generator is inherited from the parent SersicDistribution class, but will correctly use
        # this class's (SersicScaleDistribution) own sersic_func() which contains the appropriate equation for
        # a gaussian2d function.
        self.sersic_generator()

        self.normalize()


class FlatDistribution(Distribution):
    """
    Implement a source with a constant surface brightness as an ellipse, i.e. a tilted disc.

    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    minor: float
        Minor axis scale length
    major: float
        Major axis scale length
    pa: float
        Position angle in degrees of major axis measured positive in +X direction
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    Methods
    -------
    integrate_infinity():
        Normalization to flux at infinity means total flux, and that if all pixels in the
        profile are summed up, they should sum to 1.
    surface_center():
        Normalization to the center is, in this case, a no-op, because the initial
        definition of the disk sets the center (and everywhere else) to 1

    """
    def __init__(self, src, psf_upsamp):
        Distribution.__init__(self, src, psf_upsamp)
        self.norm_method = src.shape['norm_method']
        self.major=src.shape['major']
        self.minor=src.shape['minor']
        self.surf_area_units = src.shape['surf_area_units']
        self.extended = True
        self.base_center = 1.0

        self.generate()

    def generate(self):

        self.prof = self.grid.elliptical_mask(self.major, self.minor, pa=self.pa, xoff=self.xoff, yoff=self.yoff)

        self.normalize()

    def integrate_infinity(self):
        self.norm_val = self.pixelscale() / (np.pi * self.major * self.minor)

    def surface_center(self):
        self.norm_val = self.pixelscale()


class PowerDistribution(Distribution):
    """
    Create a 2-dimensional power law distribution on the current grid. I(r) = I(0) * r**-k
    There is only one configurable parameter, k, which must be positive.
    It has the same basic shape as a power law function, apart from a flat core within a radius r_core, which sidesteps
    the singularity at r=0.

    This function cannot be integrated to infinity unless k > 2. Because there is a scientific desire to have powers
    between 0 and 2, we have opted to not allow integration to infinity at all.

    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    power_index: float
        The slope of the power law index. Must be positive.
    r_core: float
        The radius of the flat core of the power law profile (arcsec)
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    Methods
    -------
    surface_center():
        Normalization is to the surface brightness of the center of the profile. This
        formulation is already normalized at that point.
    """

    def __init__(self, src, psf_upsamp):
        Distribution.__init__(self, src, psf_upsamp)
        self.power_index = src.shape['power_index']
        self.surf_area_units = src.shape['surf_area_units']
        self.r_core = src.shape['r_core']
        if self.power_index <= 0:
            raise ValueError('Power Law Index must be positive, not {}'.format(self.power_index))
        self.extended = True
        self.base_center = 1.0

        self.generate()

    def generate(self):

        self.prof = self.exponential_func(self.yrot, self.xrot, self.r_core, self.power_index)

        self.normalize()

    #def integrate_infinity(self):
    #    # integrate the Power Law profile to get the total flux for normalization, including flux outside the FOV
    #    # This is two integrations: One over the central circle; one for the exponential profile outside of that.
    #    integral = np.pi * self.r_core**2 + 2* np.pi * self.r_core**2/(self.power_index - 2)
    #    self.norm_prof = self.prof / integral * self.pixelscale()

    def surface_center(self):
        # the profile comes pre-normalized to 1 at the center.
        self.norm_val = self.pixelscale()

    def exponential_func(self, y, x, r_core, index):
        """
        Implement an exponential function. This isn't a pure exponential, because the real one is infinite at dist=0
        but this shares the same basic properties.

        Any changes to this profile will need to be made to ui/client/js/scenepage.js as well.

        Parameters
        ----------
        y: float or numpy.ndarray
            Y values for evaluating function
        x: float or numpy.ndarray
            X values for evaluating function
        r_core: float
            Radius of the flat core
        index: float
            power law index
        Returns
        -------
        profile: float or numpy.ndarray
            Float or array containing evaluated Sersic profile
        """
        dist = np.sqrt((x/r_core)**2.0 + (y/r_core)**2.0)
        profile = (dist.clip(MIN_CLIP, np.max(dist)))**(-1*index)
        # flatten the central portion. Everything within the core radius is set to 1.
        profile[np.where(dist <= 1.0)] = 1.0

        return profile



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .custom_exceptions import EngineInputError
from .utils import recursive_subclasses, merge_data

# need to import the Instrument subclasses from whereever they're defined
from .exposure import ExposureSpec
from .exposure import ExposureSpec_MultiAccum, ExposureSpec_H2RG, ExposureSpec_SiAs
from .exposure import ExposureSpec_UnevenMultiAccum, ExposureSpec_H4RG
from .exposure import ExposureSpec_SingleAccum, ExposureSpec_CCD, ExposureSpec_H1R, ExposureSpec_MAMA, ExposureSpec_XDL

def ExposureFactory(det_type, webapp=False, config={}, **kwargs):
    """
    Function to take a detector type and configuration data and build/return an appropriately
    configured instance of the desired Exposure class.

    Parameters
    ----------
    det_type: str
        The type of detector we want to make exposures with
    webapp: bool
        Switch to toggle strict API checking
    config: dict
        Configuration data segments needed to set up the exposure calculation
    **kwargs: keyword/value pairs
        Additional configuration data
    """
    method = None
    all_config = merge_data(config, dict(**kwargs))
    types = recursive_subclasses(ExposureSpec)
    detectors = [t.__name__.lower() for t in types]
    type_map = dict(list(zip(detectors, types)))

    exposure_string = "exposurespec_{}".format(det_type.lower())

    cls = type_map[exposure_string](webapp=webapp, config=config, **kwargs)
    return cls



# The telescope has a short name that we use internally.  It also has to be suitable to display to users.
telescope_short_name = 'hst'

# The telescope has a long name that we use in displays.
telescope_display_name = 'Hubble Space Telescope'

# we must list all the instruments that are present in this telescope.
# The package that supports each instrument is listed as a string so that
# we do not need to import it right now.  This is about avoiding circular
# imports.

instruments = [
    ( 'acs',        __name__ + '.acs' ),
    ( 'cos',        __name__ + '.cos' ),
    ( 'stis',       __name__ + '.stis' ),
    ( 'wfc3ir',     __name__ + '.wfc3ir' ),
    ( 'wfc3uvis',   __name__ + '.wfc3uvis' ),
]




# Licensed under a 3-clause BSD style license - see LICENSE.rst

import copy
import numpy as np
from warnings import warn

from . import observation
from .report import ReportFactory
from .scene import Scene
from .calc_utils import build_empty_scene
from .custom_exceptions import EngineInputError
from .instrument_factory import InstrumentFactory
from .strategy import StrategyFactory
from . import debug_utils
from .signal import DetectorSignal, CombinedSignal, CalculationConfig
from .noise import DetectorNoise

# pyfftw.interfaces.cache.enable()

def setup(calc_input, webapp=False, validate=False):
    """
    This generic portion of the code unpacks the input dictionary and uses the 
    pieces to configure the class instances that will perform the calculation.

    calc_input: dict
        Formatted hierarchical calculation input dictionary
    webapp: bool
        Determines whether extra API checks are done
    validate: bool
        Keyword controlling whether intermediate products are saved for validation.

    """

    warnings = {}
    try:
        scene_configuration = calc_input['scene']
        background = calc_input['background']
        background_level = calc_input.get('background_level', 'medium')
        instrument_configuration = calc_input['configuration']
        strategy_configuration = calc_input['strategy']
        if calc_input.get('debugarrays'):
            debug_utils.init(calc_input.get('debugarrays'))
    except KeyError as e:
        message = "Missing information required for the calculation: %s" % str(e)
        raise EngineInputError(value=message)

    # get the calculation configuration from the input or use the defaults
    if 'calculation' in calc_input:
        calc_config = CalculationConfig(config=calc_input['calculation'])
    else:
        calc_config = CalculationConfig()

    # #### BEGIN calculation #### #
    """
    This section currently implements the Pandeia engine API.  As the engine's object model
    is refactored, this section will have to change accordingly.
    """
    instrument = InstrumentFactory(config=instrument_configuration, webapp=webapp)
    warnings.update(instrument.warnings)

    # check for empty scene configuration and set it up properly if it is empty.
    if len(scene_configuration) == 0:
        scene_configuration = build_empty_scene()

    strategy = StrategyFactory(instrument, config=strategy_configuration, webapp=webapp, validate=validate)

    return calc_config, instrument, strategy, scene_configuration, background, background_level, warnings

def make_observation(calc_config, instrument, strategy, scene, background, background_level, webapp=False, validate=False):
    # set up the observation and then do S/N calculation...
    obs = observation.Observation(
        scene=scene,
        instrument=instrument,
        strategy=strategy,
        background=background,
        background_level=background_level,
        webapp=webapp
    )

    # seed the random number generator
    seed = obs.get_random_seed()
    np.random.seed(seed=seed)

    # Sometimes there is more than one exposure involved so implement lists for signal and noise
    my_detector_signal_list = []
    my_detector_noise_list = []
    my_detector_saturation_list = []

    if hasattr(strategy, 'dithers'):
        dither_list = strategy.dithers
    else:
        dither_list = [{'x': 0.0, 'y': 0.0}]

    for dither in dither_list:
        o = copy.deepcopy(obs)
        o.scene.offset(dither)
        # Calculate the signal rate in the detector plane. If they're configured, need to loop through
        # configured orders to include all dispersed signal.
        if instrument.projection_type in ['slitless', 'multiorder']:
            if 'orders' in instrument.disperser_config[instrument.instrument['disperser']]:
                if instrument.instrument['filter'] in instrument.disperser_config[instrument.instrument['disperser']]['orders']:
                    orders = instrument.disperser_config[instrument.instrument['disperser']]['orders'][instrument.instrument['filter']]
                else:
                    orders = instrument.disperser_config[instrument.instrument['disperser']]['orders']
            else:
                orders = None
            if 'background_orders' in instrument.disperser_config[instrument.instrument['disperser']]:
                if instrument.instrument['filter'] in instrument.disperser_config[instrument.instrument['disperser']]['background_orders']:
                    background_orders = instrument.disperser_config[instrument.instrument['disperser']]['background_orders'][instrument.instrument['filter']]
                else:
                    background_orders = instrument.disperser_config[instrument.instrument['disperser']]['background_orders']
                background_orders = instrument.disperser_config[instrument.instrument['disperser']][
                    'background_orders']
            else:
                background_orders = []
        else:
            orders = None

        if orders is not None:
            order_signals = []
            # First, take care of the orders for which a signal is desired.
            for order in orders:
                order_signals.append(DetectorSignal(o, calc_config=calc_config, webapp=webapp, validate=validate, order=order))
            # Then loop through a list of orders for which we only want to add the background calculation
            if background_orders is not None:
                o_empty = copy.deepcopy(obs)
                o_empty.scene = Scene(instrument.telescope.tel_name, config=build_empty_scene(), webapp=webapp)
                for order in background_orders:
                    order_signals.append(
                        DetectorSignal(o_empty, calc_config=calc_config, webapp=webapp, validate=validate, order=order,
                                        empty_scene=True))

            my_detector_signal = CombinedSignal(order_signals)
        else:
            my_detector_signal = DetectorSignal(o, calc_config=calc_config, webapp=webapp, validate=validate, order=None)

        my_detector_noise = DetectorNoise(my_detector_signal, o, validate=validate)

        # Every dither has a saturation map
        my_detector_saturation = my_detector_signal.get_saturation_mask()

        my_detector_signal_list.append(my_detector_signal)
        my_detector_noise_list.append(my_detector_noise)
        my_detector_saturation_list.append(my_detector_saturation)

    return my_detector_signal_list, my_detector_noise_list, my_detector_saturation_list, obs

def configure_coronagraphy(calc_input, strategy, scene_configuration, warnings, webapp=False):
    """
    Configure the engine with the various pieces needed for a coronagraphy contrast calculation.
    """

    strategy.do_contrast = False

    # Check for user-specified dithers (the contrast calculation will add an additional 2 fictional dithers).
    if not hasattr(strategy, 'dithers') or len(strategy.dithers) != 1:
        message = "Contrast calculations currently require a single dither " \
                  "to be passed in the strategy, {} was passed".format(strategy.dithers)
        raise EngineInputError(value=message)

    psf_subtraction_configuration = calc_input['strategy']['psf_subtraction_source']
    psf_subtraction_xy = strategy.psf_subtraction_xy
    pointing_error = strategy.pointing_error
    psf_subtraction_configuration['position']['x_offset'] = strategy.psf_subtraction_xy[0]
    psf_subtraction_configuration['position']['y_offset'] = strategy.psf_subtraction_xy[1]

    # add the psf_reference to the scene
    scene_configuration.append(psf_subtraction_configuration)
    scene = Scene(strategy.instrument.telescope.tel_name, config=scene_configuration, webapp=webapp)
    if hasattr(strategy, "scene_rotation"):
        scene.sources = strategy.rotate(scene.sources)
    warnings.update(scene.warnings)
    # add an attribute specifying the PSF subtraction source
    scene.psf_subtraction_idx = len(scene_configuration) - 1

    # Add the appropriate dithers for the PSF reference star and the unocculted dither
    psf_subtraction_dither = {
        'x': -strategy.psf_subtraction_xy[0] - strategy.pointing_error[0],
        'y': -strategy.psf_subtraction_xy[1] - strategy.pointing_error[1]
    }
    unocculted_dither = {'x': strategy.unocculted_xy[0], 'y': strategy.unocculted_xy[1]}

    strategy.dithers.append(psf_subtraction_dither)
    strategy.dithers.append(unocculted_dither)
    strategy.on_target = [True, False, False]

    return scene, strategy

def calculate_sn(calc_input, webapp=False, validate=False):
    """
    This is a function to do the 'forward' exposure time calculation where, given a dict
    in engine API input format, we calculate the resulting Signal/Noise and return a Report
    on the results.

    Parameters
    ----------
    calc_input: dict
        Engine API format dictionary containing the information required to perform the calculation.
    webapp: bool
        Keyword activating strict checking of API items
    validate: bool
        Keyword controlling whether intermediate validation products are saved for validation.

    Returns
    -------
    report.Report instance
    """

    calc_config, instrument, strategy, scene_configuration, background, background_level, warnings = setup(calc_input, webapp=webapp, validate=validate)

    # strategies can have different figures of interest that need to be calculated.
    # in most cases, S/N is what is desired. However, for coronagraphy the figure of interest
    # is sometimes the contrast that can be achieved. In this case, strategy.calc_type will be
    # set to 'contrast' and we need to run calculate_contrast.
    contrast = False
    if hasattr(strategy, "calc_type"):
        if strategy.calc_type == "contrast":
            contrast = True
        else:
            msg = "Unsupported calculation type: %s" % strategy.calc_type
            raise EngineInputError(value=msg)

    if contrast:
        scene, strategy = configure_coronagraphy(calc_input, strategy, scene_configuration, warnings, webapp=webapp)
    else:
        scene = Scene(instrument.telescope.tel_name, config=scene_configuration, webapp=webapp)
        warnings.update(scene.warnings)

    my_detector_signal_list, my_detector_noise_list, my_detector_saturation_list, obs = make_observation(calc_config, instrument, 
                                                                       strategy, scene, background, background_level, webapp=webapp, validate = validate)
    if strategy.method=="taspec":
        # The TASpec strategy uses one DetectorSignal and DetectorNoise, but 
        # multiple extraction boxes. We need to make copies of the DetectorSignals.
        my_detector_signal_list *= strategy.num_segments
        my_detector_noise_list *= strategy.num_segments
        my_detector_saturation_list *= strategy.num_segments

    # Use the strategy to get the extracted signal/noise products
    extracted_sn_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
    warnings.update(extracted_sn_list[0]['warnings'])


    # selection logic: If calculation_config specifies it, do that. If not, use the predefined data value
    scatter = instrument.the_detector.scatter
    if calc_config.noise['scatter'] is not None:
        scatter = calc_config.noise['scatter']
    if scatter:
        extracted_sn_list = instrument.apply_scattering(extracted_sn_list)

    if contrast:
        extracted_sn_list, warnings = calculate_contrast(my_detector_signal_list, my_detector_noise_list, extracted_sn_list, strategy, warnings, webapp=webapp)

    # #### END calculation #### #
    r = ReportFactory(calc_input, my_detector_signal_list, my_detector_noise_list, my_detector_saturation_list,
                        extracted_sn_list, warnings)

    return r


def calculate_contrast(my_detector_signal_list, my_detector_noise_list, extracted_sn_list, strategy, warnings, webapp=False):
    """
    This is a function to calculate the resulting contrast from the coronagraphic imaging calculation. 

    While this method is meant for coronagraphic modes, it will work also for regular imaging modes.

    Parameters
    ----------
    my_detector_signal_list: list
        List of DetectorSignal instances
    my_detector_noise_list: list
        List of DetectorNoise instances
    extracted_sn_list: list
        List of Strategy extraction dictionary outputs
    strategy: Strategy instance
        A configured Strategy instance for the contrast calculation
    warnings: dict
        A dictionary of warnings produced within all preceeding code.
    webapp: bool
        Control whether we are running in webapp (strict API) mode, or not.

    Returns
    -------
    report.Report instance
    """

    # We need a regular S/N of the target source
    extracted_sn_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
    warnings.update(extracted_sn_list[0]['warnings'])

    # Use the strategy to get the extracted contrast products
    grid = my_detector_signal_list[0].grid

    aperture = strategy.aperture_size
    annulus = strategy.__dict__.get("sky_annulus", (0,0))

    # Create a list of contrast separations for which to calculate the contrast
    bounds = grid.bounds()
    ncontrast = strategy.ncontrast
    contrasts = np.zeros(ncontrast)
    contrast_separations = np.linspace(0 + aperture, bounds['xmax'] - annulus[1], ncontrast)
    contrast_azimuth = np.radians(strategy.contrast_azimuth)
    contrast_xys = [(separation * np.sin(contrast_azimuth),
                     separation * np.cos(contrast_azimuth)) for separation in contrast_separations]

    # Calculate contrast at each separation
    for i, contrast_xy in enumerate(contrast_xys):
        strategy.target_xy = contrast_xy
        extracted_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
        contrasts[i] = extracted_list[0]['extracted_noise']

    # What is the flux of the unocculted star.
    # We set the do_contrast attribute to True so the unocculted dither will be used.
    strategy.do_contrast = True
    strategy.on_target = [False, False, True]
    strategy.target_xy = strategy.unocculted_xy
    extract_unocculted_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
    extracted_sn_list.append(extract_unocculted_list[0])

    # Contrast is relative to the unocculted on-axis star.
    contrasts /= extract_unocculted_list[0]['extracted_flux']
    contrast_curve = [contrast_separations, contrasts]

    # #### END calculation #### #

    # Add the contrast curve and link relevant saturation maps to the extracted_sn dict for passing
    for x in range(len(extracted_sn_list)):
        extracted_sn_list[x]['contrast_curve'] = contrast_curve

    return extracted_sn_list, warnings

def calculate_exposure_time(calc_input, webapp=False, validate=False, **kwargs):
    """
    This is a function to do the 'reverse' exposure time calculation where given a desired
    signal-to-noise ratio we calculate the optimal exposureSpecification and return a Report
    on the results.

    Parameters
    ----------
    calc_input: dict
        Dictionary containing the information required to perform the calculation.
    webapp: bool
        Keyword activating strict checking of API items
    validate: bool
        Keyword controlling whether intermediate products are saved for validation.

    Returns
    -------
    report.Report instance
    """
    calc_config, instrument, strategy, scene_configuration, background, background_level, warnings = setup(calc_input, webapp=webapp, validate=validate)
    # do a forward calculation with time=1
    # this sets the time in the calculation 
    instrument.the_detector.set_time(1.0)

    # strategies can have different figures of interest that need to be calculated.
    # in most cases, S/N is what is desired. However, for coronagraphy the figure of interest
    # is sometimes the contrast that can be achieved. In this case, strategy.calc_type will be
    # set to 'contrast' and we need to run calculate_contrast.
    contrast = False
    if hasattr(strategy, "calc_type"):
        if strategy.calc_type == "contrast":
            contrast = True
        else:
            msg = "Unsupported calculation type: %s" % strategy.calc_type
            raise EngineInputError(value=msg)

    if contrast:
        scene, strategy = configure_coronagraphy(calc_input, strategy, scene_configuration, warnings, webapp=webapp)
    else:
        scene = Scene(instrument.telescope.tel_name, config=scene_configuration,  webapp=webapp)
        warnings.update(scene.warnings)

    my_detector_signal_list, my_detector_noise_list, my_detector_saturation_list, obs = make_observation(calc_config, instrument, 
                                                                       strategy, scene, background, background_level, webapp=webapp, validate=validate)

    if strategy.method=="taspec":
        # The TASpec strategy uses one DetectorSignal and DetectorNoise, but multiple
        # extraction boxes. We need to make copies of the DetectorSignals.
        my_detector_signal_list = [my_detector_signal_list[0]] * strategy.num_segments
        my_detector_noise_list = [my_detector_noise_list[0]] * strategy.num_segments
        my_detector_saturation_list = [my_detector_saturation_list[0]] * strategy.num_segments

    # Use the strategy to get the extracted signal/noise products
    extracted_sn_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
    warnings.update(extracted_sn_list[0]['warnings'])

    # Detector Scattering.  If calculation_config specifies to apply it, do that. If not,
    # use the predefined data value.
    # 
    # This detector effect is unusual for two reasons: One, it's defined as a fraction of
    # the rate, but must be added to the rate_plus_bg; Two, it's only present in the
    # noise.
    scatter = instrument.the_detector.scatter
    if calc_config.noise['scatter'] is not None:
        scatter = calc_config.noise['scatter']
    if scatter:
        extracted_sn_list = instrument.apply_scattering(extracted_sn_list)

    if strategy.method == "taspec":
        # TASpec is unique amongst strategies; it pulls up to 3 results off a single
        # DetectorSignal. However, to avoid duplicating the DetectorSignal, we employ a
        # special variable unique to TASpec to loop through the segments, not the signals.
        signal = 0
        new_extracted_sn_list = []
        for segment in range(strategy.num_segments):
            strategy.segment_number = segment
            new_time = my_detector_signal_list[signal].the_detector.scale_exposure_time(extracted_sn_list[segment], calc_input)
            # reset the time(s)
            obs.instrument.the_detector.set_time(new_time)
            my_detector_signal_list[signal].the_detector.set_time(new_time)
            my_detector_signal_list[signal].instrument.the_detector.set_time(new_time)
        
            # retrigger calculations using the new time
            my_detector_signal_list[signal].signal_products()
            my_detector_noise_list[signal] = DetectorNoise(my_detector_signal_list[signal], obs, validate=validate)
            my_detector_saturation_list[signal] = my_detector_signal_list[signal].get_saturation_mask()
            new_extracted_sn_list.extend(strategy.extract([my_detector_signal_list[signal]], [my_detector_noise_list[signal]]))
        # normally, the Strategy would put the combined signal first.
        new_extracted_sn_list.insert(0, new_extracted_sn_list[-1])
        extracted_sn_list = new_extracted_sn_list
    else:
        for signal in range(len(my_detector_signal_list)):
            new_time = my_detector_signal_list[signal].the_detector.scale_exposure_time(extracted_sn_list[signal], calc_input)
            # reset the time(s)
            obs.instrument.the_detector.set_time(new_time)
            my_detector_signal_list[signal].the_detector.set_time(new_time)
            my_detector_signal_list[signal].instrument.the_detector.set_time(new_time)
        
            # retrigger calculations using the new time
            my_detector_signal_list[signal].signal_products()
            my_detector_noise_list[signal] = DetectorNoise(my_detector_signal_list[signal], obs, validate=validate)
            my_detector_saturation_list[signal] = my_detector_signal_list[signal].get_saturation_mask()

        # Use the strategy to get the extracted signal/noise products
        extracted_sn_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)

    if scatter:
        extracted_sn_list = instrument.apply_scattering(extracted_sn_list)

    if contrast:
        extracted_sn_list, warnings = calculate_contrast(my_detector_signal_list, my_detector_noise_list, extracted_sn_list, strategy, warnings, webapp=webapp)

    # #### END calculation #### #
    r = ReportFactory(calc_input, my_detector_signal_list, my_detector_noise_list, my_detector_saturation_list,
                        extracted_sn_list, warnings)


    return r



"""
This files contains string constants which are reserved for use within the code

"""

#these are the possible name values for collections
#and items in them
TARGET = "target"
READNOISE = "read_noise"
POST_FLASH = "post_flash"
SKY = "sky"
STRAY_LIGHT = "stray_light"
TARGET_SCATTERED_LIGHT = "target_scattered_light"
DARK = "dark"
BACKGROUND = "background"
THERMAL = "thermal"
TOTAL = "total"
BRIGHTEST_PIXEL = "brightest_pixel"
DETECT_DARK = "detect_dark"
DARK_ONLY = 'dark_only'
LINE_BROADENED = "line_broadened"
LINE_DROPPED = "line_dropped"

#these are different types of region shapes
SQUARE = 'square'
CIRCLE = 'circle'
PERCENT = 'percent'
RECTANGLE = 'rectangle'

#units
PIXELS = 'pixels'
ARCSEC = 'arcsec'

#these are different types of regions
SNR_REGION = 'snr_region'
ACQ_WINDOW = 'acquisition_window'
PIX_REGION = 'region_height_by_pixel_width'
AREA_WHOLE_DETECTOR = 'area_over_whole_detector'
AREA_PIXEL = 'area_over_one_pixel'

#simulation modes
CALCULATING_SNR = 'SNR'
CALCULATING_TIME = 'Time'

# these are different types of slit/aperture kinds
FILTERED = "filtered"
LOSSY = "lossy"
GEOMETRIC = "geometric" # place holder (not explicitly used by code)

# instrument types and modes
IMAGER = "imager"
SPECTROGRAPH = "spectrograph"
IMAGING = "imaging"
SCIMAGING = "scimaging"   # scan mode imaging
RAMP_FILTER = "rampfilter"
TARGET_ACQUISITION = "targetacquisition"
SPECTROSCOPIC = "spectroscopic"
SCSPECTROSCOPIC = "scspectroscopic"
SPECTROSCOPIC_ACQ = "spectroscopicacq"

# Common warning texts
PARTIAL_OVERLAP_KEY = "PartialOverlap"
PARTIAL_OVERLAP_MESSAGE = "Partial overlap between instrument throughput band and input spectrum. Please doublecheck the wavelength range covered by the selected spectrum, and the instrument documentation for the selected filter."
PARTIAL_TARGET_OVERLAP_MESSAGE = PARTIAL_OVERLAP_MESSAGE.replace("input spectrum", "target spectrum")
PARTIAL_SKY_OVERLAP_MESSAGE = PARTIAL_OVERLAP_MESSAGE.replace("input spectrum", "sky spectrum")
ABMAG_KEY = "Normalization"
ABMAG_MESSAGE = "Vegamag selected while using Galex/Sloan normalization passband."






"""
Function signatures:
--------------------

``compatibility(engine_input, result)``
    No return value. 'result' is the engine_input after compatibility changes are applied.

``get_configuration_data(ei)``
    Returns instrument configuration dictionaries given the engine input dictionary.

``get_detector_tables(instrument_name, detector_name, science_mode, spec_element)``
    Returns a dictionary with the enclosed energy tables associated with the disperser.

``modify_areas_of_interest(instrument)``
    No return value; modifies instrument.areas_of_interest

``make_snr_region(instrument, *args, **kwargs)`` #different for imager or spectrograph
    Returns an AreaOverWhich of the correct type; modifies instrument.areas_of_interest

``make_area_pixel(instrument, *args, **kwargs)`` #different for imager or spectrograph?
    Returns an AreaOverWhich for 1-pix extraction; modifies instrument.areas_of_interest

``modify_configuration(instrument, custom_config)``
    No return value. Modifies the Instrument instance in place.

``modify_target_or_sky(instrument, observed_target, observed_sky, obswave=None)``
    Returns new instances of ObservedItem with the modified target and
    sky, or references to the input, if no modification took place.

``stray_light(instrument, target, sky)``
    Returns a ObservedItem with the stray light component, or None.

``make_readpattern(instrument, engine_input)``
    Returns a ReadPattern instance configured for the instrument and user inputs.

``modify_noise(noise_collection)``
    No return value. Modifies the NoiseCollection instance in place.

``modify_total(instrument, total_collection)``
    No return value. Modifies the total DetectedCollection instance in place.

``modify_one_pixel(instrument, extracted_collection)``
    Returns a modified copy of the input one-pixel ExtractedCollection.

``detect_dark(instrument, dark)``
    Returns a DetectedItem or DetectedCollection with the value of the dark current per pixel.

``prepare_vectors_for_plotting(observed_target, background)``
    Modifies target and/or background vector objects according to
    the expectations of what the plots and table should do.
"""

from .telescopes import TELESCOPES
#from ..main.tools import find_instrument
from pandeia.engine.helpers.bit.pyetc_util import dynamic_import, stable_dict, find_instrument

MODULE_PATH_NAME = 'pandeia.engine.helpers.bit.instruments.%s.%s.instr_functions'

# Custom configuration functions handle instrument-specific configuration data
#
ALLOWED_CUSTOM_CONFIG = ['compatibility',
                         'get_configuration_data',
                         'modify_configuration'
                        ]

# Custom calculation functions implement instrument-specific calculations.
#
# Although the detector tables are read as configuration information,
# they actually represent PSF calculations that are expressed via table
# interpolation. Hence their inclusion in the subset of calculation functions.
ALLOWED_CUSTOM_CALC = ['get_detector_tables',
                       'make_snr_region',
                       'modify_areas_of_interest',
                       'modify_target_or_sky',
                       'stray_light',
                       'make_area_whole_detector',
                       'make_area_pixel',
                       '_init_custom_config',
                       'compute_buffer_time',
                       'make_readpattern',
                       'modify_total',
                       'modify_noise',
                       'modify_one_pixel',
                       'detect_dark',
                       'prepare_vectors_for_plotting']

def import_custom_overrides(instrument_name):
    """ Imports and makes available to the caller a dictionary of
        instrument-specific functions.

    Parameters
    ----------

    instrument_name: str
       name of instrument

    Returns
    -------

    result: dict
        dict with instrument-specific functions.

    """
    result = stable_dict()

#    try:
#        instr_module = find_instrument(TELESCOPES[instrument_name], instrument_name + ".instr_functions")

        # workaround for the above
    module_name = MODULE_PATH_NAME % (TELESCOPES[instrument_name], instrument_name)
    instr_module = dynamic_import(module_name)
#    except ImportError, e:
#        # No custom imports
#        return result

    for key in ALLOWED_CUSTOM_CONFIG:
        try:
            result[key] = instr_module.__getattribute__(key)
        except AttributeError as e:
            continue

    for key in ALLOWED_CUSTOM_CALC:
        try:
            result[key] = instr_module.__getattribute__(key)
        except AttributeError as e:
            continue

    return result




# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

from .config import DefaultConfig
from .custom_exceptions import DataConfigurationError, EngineInputError, DataError
from .pandeia_warnings import source_warning_messages as warning_messages
from .utils import merge_data

from . import config as cf
from . import io_utils as io

default_refdata_directory = cf.default_refdata_directory


class Line(DefaultConfig):

    """
    Container for spectral line parameters

    Parameters
    ----------
    config: dict (optional)
        dictionary containing necessary configuration information
    **kwargs: list of keyword/value pairs
        parameter keyword and value pairs to augment defaults and config

    Attributes
    ----------
    id: str or int
        Line ID tag
    name: str
        Name given to the line
    center: float
        Central wavelength of the line in microns
    width: float
        Line FWHM in km/s
    strength: float
        Line strength in erg/cm^2/s for emission or central optical depth for absorption
    profile: str
        Line profile change, currently only "gaussian" is supported
    emission_or_absorption: str
        Line profile type, either 'emission' or 'absorption'
    """
    def _get_config(self, webapp=False):
        """
        Read default configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        self.webapp = webapp
        # use this trick to key the configuration file name off the name of the instantiated subclass
        ref_dir = os.path.join(default_refdata_directory, "source", "line")
        if self.webapp:
            config = {}
        else:
            config = io.read_json(os.path.join(ref_dir, "defaults.json"), raise_except=True)
        return config


class Source(DefaultConfig):

    """
    Container for source parameters

    Parameters
    ----------
    config: dict (optional)
        dictionary containing necessary configuration information
    telescope: str
        The name of the telescope the calculation is being computed for
    **kwargs: list of keyword/value pairs
        parameter keyword and value pairs to augment defaults and config

    Attributes
    ----------
    id: str or int
        source id
    position: dict
        x_offset: float
            x offset (arcsec)
        y_offset: float
            y offset (arcsec)
        orientation: float
            orientation angle (degrees)
    shape: dict
        geometry: string
            supported source geometries are:
                "point" - requires no further parameters

                "flat" - requires the following parameters:
                    major: float
                        major axis length (arcsec).
                    minor: float
                        minor axis length (arcsec).

                "gaussian2d" - requires "major" and "minor"

                "sersic" - requires "major", "minor", and an additional parameter:
                    sersic_index: float
                        power law index that sets the shape of a sersic profile. used only for "sersic" geometry.
                            sersic_index = 1.0 --> exponential
                            sersic_index = 0.5 --> gaussian
                            sersic_index = 4.0 --> de Vaucouleurs
    spectrum: dict
        lines: list
            list of line definitions
        name: string
            Name of the source
        redshift: float
            Redshift of the source
        sed: dict
          Defines the spectral energy distribution of the spectrum.

            sed_type: string
                Type of the spectral energy distribution. Each type requires its own set
                of parameters. The analytic sed_type's (none, flat, powerlaw, flat) all
                require 'wmin', 'wmax', and 'sampling' to define the range and wavelength
                sampling over which the model spectrum is calculated. However, they are only
                available in the API for testing purposes and should not be configured via
                the UI.

                    no_continuum - No continuum, specifically F(lam) = 0.0 over specified range [wmin, wmax]
                        wmin: float (default 0.5)
                            Minimum wavelength in microns
                        wmax: float (default 30.0)
                            Maximum wavelength in microns
                        sampling: int (default 200)
                            Sets the logarithmic wavelength sampling of the model spectrum

                    flat - Flat spectrum in specified units calculated over specified range [wmin, wmax]
                        wmin: float (default 0.5)
                            Minimum wavelength in microns
                        wmax: float (default 30.0)
                            Maximum wavelength in microns
                        sampling: int (default 200)
                            Sets the logarithmic wavelength sampling of the model spectrum
                        unit: string
                            Units of spectrum, either 'fnu' or 'flam'

                    powerlaw - Powerlaw spectrum where F ~ lambda ^ index calculated over range [wmin, wmax]
                        wmin: float (default 0.5)
                            Minimum wavelength in microns
                        wmax: float (default 30.0)
                            Maximum wavelength in microns
                        sampling: int (default 200)
                            Sets the logarithmic wavelength sampling of the model spectrum
                        unit: string
                            Units of spectrum, either 'fnu' or 'flam'
                        index: float
                            Exponent of the power law

                    blackbody - Blackbody spectrym calculated over range [wmin, wmax]
                        wmin: float (default 0.5)
                            Minimum wavelength in microns
                        wmax: float (default 30.0)
                            Maximum wavelength in microns
                        sampling: int (default 200)
                            Sets the logarithmic wavelength sampling of the model spectrum
                        temp: float
                            Temperature of the blackbody

                    phoenix - Parameterized stellar atmosphere models calculated by the Phoenix group
                        key: string
                            In webapp mode, a key is used to look up a predefined set of parameters. If not
                            in webapp mode and if key is not provided, model parameters can be passed directly:
                        teff: float
                            Effective temperature. Allowed range is 2000 K to 70000 K
                        log_g: float
                            Logarithm of the surface gravity in cgs units. Allowed range is 0.0 to 5.5.
                        metallicity: float
                            Logarithm of the metallicity in units of solar metallicity. Allowed range is -4.0 to 0.5.

                    hst_calspec - HST standard star spectra
                        key: string
                            Key used to look up which spectrum load.

                    galaxies - Integrated spectra of galaxies from Brown et al. (2014)
                        key: string
                            Key used to look up which spectrum load.

                    input - spectrum provided via input arrays
                        spectrum: list-like or numpy.ndarray
                            The 0th index is taken to be wavelength in units of 'microns'.
                            The 1st index is taken to be the flux in units of 'mJy'.

        normalization: dict
            Define the brightness of the source
                type: string
                    Method of normalization to perform.
                    Supported methods are: "at_lambda", "hst", "jwst", "photsys", and "none"
                norm_wave: float
                    Reference wavelength in 'norm_waveunit' at which spectrum will be scaled for type 'at_lambda'.
                norm_flux: float
                    Reference flux in 'norm_fluxunit' to which spectrum will be scaled.
                norm_fluxunit: string
                    Specify the flux units in which the normalization should occur.
                    Supports flam, fnu, vegamag, abmag, mjy, ujy, njy, jy
                norm_waveunit: string
                    Specify the wavelength units used in normalization
                bandpass: string
                    Specifies how bandpass information is looked up or calculated for normalization
                    types 'hst', 'jwst', and 'photsys'.

    Methods
    -------
    add_line, remove_line
    """
    def __init__(self, telescope, config={}, webapp=False, **kwargs):
        """
        How a source is configured depends on the defaults, the shape-specific defaults, and any input parameters.

        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        telescope: str
            The telescope being used (not formally a part of the engine input API)
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        self.webapp = webapp
        self.telescope = telescope
        DefaultConfig.__init__(self, webapp=webapp, config=config, **kwargs)

        if self.shape['geometry'] in self.geometries:
            shape_config = self.geometries[self.shape['geometry']]
            shape_config = self._load_dict(shape_config, "required", msg="Missing required shape data.")
            if self.webapp:
                shape_config = self._load_dict(shape_config, "api", msg="Missing required shape API data.")
                shape_config.pop("defaults", None)
            else:
                shape_config = self._load_dict(shape_config, "defaults", msg="Missing default shape configuration.")
                shape_config.pop("api", None)
            # clean out stuff not needed by engine
            shape_config.pop("display_string")
            self.shape = merge_data(shape_config, self.shape)
        else:
            msg = "Unsupported source geometry: %s" % self.shape['geometry']
            raise EngineInputError(value=msg)

        all_config = merge_data(config, dict(**kwargs))

        if self.webapp:
            self._nested_api_checks(self.__dict__)

    def _get_config(self):
        """
        Read default configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # use this trick to key the configuration file name off the name of the instantiated subclass
        try:
            ref_dir = os.path.join(default_refdata_directory, self.telescope, "source")
            config = io.read_json(os.path.join(ref_dir, "config.json"), raise_except=True)
        except (OSError, TypeError):
            raise DataError(f"Source configuration requires a valid telescope to be specified, got {self.telescope}.")

        if self.webapp:
            self._load_dict(config, "api", msg="No default source configuration found")
            config.pop("defaults")
        else:
            config = self._load_dict(config, "defaults", msg="No default source configuration found")
            config.pop("api")

        return config

    def _nested_api_checks(self, conf):
        """
        If webapp=True, check to make sure all of the configured shape_parameters and spectrum_parameters exist in the input
        configuration.  api_parameters only works at top level, but we need to check one level down
        in self.shape and self.spectrum which is why this is a separate check.  Proper fix would be to refactor Source
        into sub-classes for spectrum and each geometry type.
        """
        # 'shape' and 'spectrum' are already checked for via api_parameters
        for section in ['shape', 'spectrum', 'position']:
            if section in conf:
                parskey = '%s_parameters' % section
                params = getattr(self, section)
                if parskey in getattr(self, section):
                    for p in params:
                        if p not in conf[section]:
                            key = "source_{:s}_parameter_{:s}_missing".format(section, p)
                            msg = warning_messages["source_parameter_missing"].format(
                                section,
                                p,
                                getattr(self, section)[p]
                            )
                            self.warnings[key] = msg
                else:
                    key = "source_%s_parameters_missing" % section
                    msg = warning_messages["source_parameters_missing"].format(section)
                    self.warnings[key] = msg

    def _sanity_checks(self):
        """
        Make sure normalization, SED, and shape information are defined.
        """
        # need this information to make a source...
        if 'normalization' not in self.spectrum:
            msg = "Normalization information missing for default source."
            raise DataConfigurationError(value=msg)
        if 'sed' not in self.spectrum:
            msg = "SED information missing for default source."
            raise DataConfigurationError(value=msg)
        if not hasattr(self, "shape"):
            msg = "Shape information missing for default source."
            raise DataConfigurationError(value=msg)
        # make sure configured shape_parameters exist in the data files
        if "shape_parameters" in self.shape:
            for p in self.shape['shape_parameters']:
                if p not in self.shape:
                    msg = "Shape information is missing %s for default source." % p
                    raise DataConfigurationError(value=msg)

    def add_line(self, config={}, **kwargs):
        """
        Add a spectral Line to the Source's spectrum.
        """
        # use a DefaultConfig subclass here for convenience to populate
        # defaults, but store the dict format in spectrum['lines'] to simplfify
        # instantiation from JSON
        line_definition = Line(config=config, **kwargs)
        self.warnings.update(line_definition.warnings)

        self.spectrum['lines'].append(line_definition.__dict__)

    def remove_line(self, line_definition):
        """
        Remove a given line_definition from Source's spectrum
        """
        self.spectrum['lines'].remove(line_definition)


def load_empty_source():
    """
    Special function for the special case of loading an empty source
    """
    try:
        ref_dir = os.path.join(default_refdata_directory, "source")
        config = io.read_json(os.path.join(ref_dir, "empty.json"), raise_except=True)
    except (OSError, TypeError):
        raise DataError(f"Could not load the empty source.")
    return config



'''Map the sky background form parameters into a form that the etc_engine can use:
(ie, a synphot expression)
keep this file in sync with 
https://github.com/spacetelescope/pyetc/blob/master/pyetc/etc_web/etc/sky_angles.py
'''

import os
import datetime
import numpy as np

from pandeia.engine.helpers.background.hst import util
from pandeia.engine.custom_exceptions import BackgroundError, RangeError

DAYS_IN_YEAR = 365.25
OBLIQUITY = 23.43 / 180.0 * np.pi


def radec_to_elonglat(ra, dec):
    '''convert ra, dec to ecliptic long,long (all in decimal deg)'''

    # degrees to radians
    ra = ra / 180.0 * np.pi
    dec = dec / 180.0 * np.pi

    # ecliptic latitude
    sin_lat = np.sin(dec) * np.cos(OBLIQUITY) - np.cos(dec) * np.sin(OBLIQUITY) * np.sin(ra)
    lat = np.arcsin(sin_lat)

    # ecliptic longitude
    tan_long = (np.tan(dec) * np.sin(OBLIQUITY) + np.cos(OBLIQUITY) * np.sin(ra)) / np.cos(ra)
    long = np.arctan(tan_long)

    # convert longitude to appropriate quadrant
    cos_long = np.cos(dec) * np.cos(ra) / np.cos(lat)
    sin_long= (np.sin(dec) * np.sin(OBLIQUITY) + (np.cos(dec) * np.sin(ra) * np.cos(OBLIQUITY))) / np.cos(lat)
    long_cos = np.arccos(cos_long)
    long_sin = np.arcsin(sin_long)
    if long < 0 and long_sin > 0 and long_cos > 0:
        long += np.pi
    elif long > 0 and long_sin < 0 and long_cos > 0:
        long += np.pi
    elif long < 0 and long_sin < 0 and long_cos > 0:
        long += np.pi * 2.

    # radians to degrees,
    lat  = lat * 180.0 / np.pi
    long = long * 180.0 / np.pi

    return long, lat


def radec_date_to_helonglat(ra, dec, date):
    '''compute the heliocentric ecliptic longitude/latitude given ra, dec and obs date

    This is not a high precision routine!!  (since zodical variance with angle is slow)
    Input and output angles are in degrees, date is a datetime object. Angle
    computations done in radians, however.
    '''
    # determine the longitude of the vernal equinox given the date
    # calculate fraction of year separating the dates
    veq = datetime.date(date.year,3,20)
    delta = (date - veq).days
    if delta > DAYS_IN_YEAR:
        delta -= DAYS_IN_YEAR
    sunangle = 360.*delta/DAYS_IN_YEAR
    elong, elat = radec_to_elonglat(ra, dec)

    # Don't limit the ranges of elong and sunangle (using
    # limit_longitude) until ready to look up in the table.
    return elong-sunangle, elat, sunangle

def limit_longitude(theta):
    theta = np.abs(theta)
    if theta > 180.0:
        return 360.0 - theta
    else:
        return theta

def limit_latitude(theta):
    theta = np.abs(theta)
    if theta > 90.0:
        return 180.0 - theta
    else:
        return theta

def radec_sunangle_to_helonglat(ra, dec, sunangle):
    '''compute the heliocentric ecliptic longitude/latitude given ra, dec and longitudnal
    angle to sun

    This is not a high precision routine!!  (since zodical variance with angle is slow)
    Input and output angles are in degrees, date is a datetime object. Angle
    computations done in radians, however.
    '''
    elong, elat = radec_to_elonglat(ra, dec)
    # Don't limit the ranges of elong and sunangle (using
    # limit_longitude) until ready to look up in the table.
    return sunangle, elat

def radec_to_decimal(in_ra, in_dec):
    """Convert in_ra, in_dec from input values to floating-point degrees; returns ra, dec

    For each, if we can convert it to a float, use the result.  Otherwise,
    parse HH:MM:SS or DD:MM:SS to convert it
    """

    try:
        ra = float(in_ra)
    except ValueError:
        # Then it must be hours:minutes:seconds
        l = in_ra.split(':')

        # convert to hours
        ra = int(l[0])
        if len(l) > 1 :
            ra += float(l[1])/60.
        if len(l) > 2 :
            ra += float(l[2])/3600.
        #
        # convert hours to degrees
        ra *= 15.

    try:
        dec = float(in_dec)
    except ValueError:
        # then it must be degrees:minutes:seconds
        l = in_dec.split(':')

        # get unsigned degrees
        dec = np.abs(int(l[0]))
        if len(l) > 1 :
            dec += float(l[1])/60.
        if len(l) > 2 :
            dec += float(l[2])/3600.

        # apply the sign to it.
        if in_dec.strip().startswith('-') :
            dec = -dec

    return ra, dec

def bilinear_interp(xref, yref, arr, x, y):
    '''perform bilinear interpolation on 2-d array.

    xref: contains x values corresponding to first dim index in array.
    yref: contains y values corresponding to second dim index in array.
    arr: array of values to be interpolated.
    x, y: x and y values for interpolation, scalars or arrays permitted.

    Poor error handling at the moment...
    '''

    xind = np.searchsorted(xref, x)
    yind = np.searchsorted(yref, y)

    # Bound the indices to 0 .. N-1
    xind0 = np.maximum(0, xind - 1)
    yind0 = np.maximum(0, yind - 1)
    xind1 = np.minimum(len(xref) - 1, xind)
    yind1 = np.minimum(len(yref) - 1, yind)

    # The denominator may be zero if outside of the bounds of the
    # image
    xden = (xref[xind1] - xref[xind0])
    xden = np.where(xden == 0.0, 1.0, xden)
    yden = (yref[yind1] - yref[yind0])
    yden = np.where(yden == 0.0, 1.0, yden)

    xfract = (x - xref[xind0]) / xden
    xfracti = 1.0 - xfract
    yfract = (y - yref[yind0]) / yden
    yfracti = 1.0 - yfract

    val = (arr[xind0,yind0] * xfracti * yfracti +
           arr[xind1,yind0] * xfract * yfracti +
           arr[xind0,yind1] * xfracti * yfract +
           arr[xind1,yind1] * xfract * yfract)
    return val

def zodi_lookup(helong, helat, posfile=None):
    '''look up zodical magnitude given ecliptic latitude and longitude in deg.'''
    #This lookup table presently lives with the source code
    if posfile is None:
        posfile = os.path.join(os.path.dirname(__file__),
                               'zodi.dat')
    edict = util.read_dict(posfile)

    # The table is symmetric for 0 <= long <= 180 and 0 <= lat <= 90,
    # so limit the lookup values to those ranges.
    helong = limit_longitude(helong)
    helat = limit_latitude(helat)

    table = np.array(edict['table'])
    marr = table[1:,1:] #table contents
    xref = table[1:,0]  #header (index) row
    yref = table[0,1:]  #header (index) column
    # look up value
    zmag = bilinear_interp(xref, yref, marr, np.abs(helong), np.abs(helat))

    if np.isnan(zmag):
        raise RangeError('Helio-ecliptic long and ecliptic lat (%f, %f) outside valid range of zodiacal light table.'
                         %(helong,helat))
    return zmag
